"""
Versioning: we use a hard-coded version number, because it's simple and always
works. For dev installs we add extra version info from Git.
"""

import logging
import subprocess
from pathlib import Path


# This is the reference version number, to be bumped before each release.
# The build system detects this definition when building a distribution.
__version__ = "0.5.2"

# Allow using nearly the same code in different projects
project_name = "fastplotlib"


logger = logging.getLogger(project_name.lower())

# Get whether this is a repo. If so, repo_dir is the path, otherwise repo_dir is None.
repo_dir = Path(__file__).parents[1]
repo_dir = repo_dir if repo_dir.joinpath(".git").is_dir() else None


def get_version():
    """Get the version string."""
    if repo_dir:
        return get_extended_version()
    else:
        return __version__


def get_extended_version():
    """Get an extended version string with information from git."""

    release, post, labels = get_version_info_from_git()

    # Sample first 3 parts of __version__
    base_release = ".".join(__version__.split(".")[:3])

    # Check release
    if not release:
        release = base_release
    elif release != base_release:
        logger.warning(
            f"{project_name} version from git ({release}) and __version__ ({base_release}) don't match."
        )

    # Build the total version
    version = release
    if post and post != "0":
        version += f".post{post}"
    if labels:
        version += "+" + ".".join(labels)

    return version


def get_version_info_from_git():
    """Get (release, post, labels) from Git.

    With `release` the version number from the latest tag, `post` the
    number of commits since that tag, and `labels` a tuple with the
    git-hash and optionally a dirty flag.
    """

    # Call out to Git
    command = [
        "git",
        "describe",
        "--long",
        "--always",
        "--tags",
        "--dirty",
        "--first-parent",
    ]
    try:
        p = subprocess.run(command, cwd=repo_dir, capture_output=True)
    except Exception as e:
        logger.warning(f"Could not get {project_name} version: {e}")
        p = None

    # Parse the result into parts
    if p is None:
        parts = (None, None, "unknown")
    else:
        output = p.stdout.decode(errors="ignore")
        if p.returncode:
            stderr = p.stderr.decode(errors="ignore")
            logger.warning(
                f"Could not get {project_name} version.\n\nstdout: "
                + output
                + "\n\nstderr: "
                + stderr
            )
            parts = (None, None, "unknown")
        else:
            parts = output.strip().lstrip("v").split("-")
            if len(parts) <= 2:
                # No tags (and thus also no post). Only git hash and maybe 'dirty'
                parts = (None, None, *parts)

    # Return unpacked parts
    release, post, *labels = parts
    return release, post, labels


__version__ = get_version()

version_info = tuple(
    int(i) if i.isnumeric() else i for i in __version__.split("+")[0].split(".")
)



from itertools import product
from math import ceil

import numpy as np
import pygfx

from ._base import GraphicFeature, GraphicFeatureEvent, block_reentrance

VOLUME_RENDER_MODES = {
    "mip": pygfx.VolumeMipMaterial,
    "minip": pygfx.VolumeMinipMaterial,
    "iso": pygfx.VolumeIsoMaterial,
    "slice": pygfx.VolumeSliceMaterial,
}


class TextureArrayVolume(GraphicFeature):
    """
    Manages an array of Textures representing chunks of an image. Chunk size is the GPU's max texture limit.

    Creates and manages multiple pygfx.Texture objects.
    """

    event_info_spec = [
        {
            "dict key": "key",
            "type": "slice, index, numpy-like fancy index",
            "description": "key at which image data was sliced/fancy indexed",
        },
        {
            "dict key": "value",
            "type": "np.ndarray | float",
            "description": "new data values",
        },
    ]

    def __init__(self, data, isolated_buffer: bool = True):
        super().__init__()

        data = self._fix_data(data)

        shared = pygfx.renderers.wgpu.get_shared()

        self._texture_size_limit = shared.device.limits["max-texture-dimension-3d"]

        if isolated_buffer:
            # useful if data is read-only, example: memmaps
            self._value = np.zeros(data.shape, dtype=data.dtype)
            self.value[:] = data[:]
        else:
            # user's input array is used as the buffer
            self._value = data

        # data start indices for each Texture
        self._row_indices = np.arange(
            0,
            ceil(self.value.shape[1] / self._texture_size_limit)
            * self._texture_size_limit,
            self._texture_size_limit,
        )
        self._col_indices = np.arange(
            0,
            ceil(self.value.shape[2] / self._texture_size_limit)
            * self._texture_size_limit,
            self._texture_size_limit,
        )

        self._zdim_indices = np.arange(
            0,
            ceil(self.value.shape[0] / self._texture_size_limit)
            * self._texture_size_limit,
            self._texture_size_limit,
        )

        shape = (self.zdim_indices.size, self.row_indices.size, self.col_indices.size)

        # buffer will be an array of textures
        self._buffer: np.ndarray[pygfx.Texture] = np.empty(shape=shape, dtype=object)

        self._iter = None

        # iterate through each chunk of passed `data`
        # create a pygfx.Texture from this chunk
        for _, buffer_index, data_slice in self:
            texture = pygfx.Texture(self.value[data_slice], dim=3)

            self.buffer[buffer_index] = texture

    @property
    def value(self) -> np.ndarray:
        """The full array that represents all the data within this TextureArray"""
        return self._value

    def set_value(self, graphic, value):
        self[:] = value

    @property
    def buffer(self) -> np.ndarray[pygfx.Texture]:
        """array of buffers that are mapped to the GPU"""
        return self._buffer

    @property
    def row_indices(self) -> np.ndarray:
        """
        row indices that are used to chunk the big data array
        into individual Textures on the GPU
        """
        return self._row_indices

    @property
    def col_indices(self) -> np.ndarray:
        """
        column indices that are used to chunk the big data array
        into individual Textures on the GPU
        """
        return self._col_indices

    @property
    def zdim_indices(self) -> np.ndarray:
        """
        z dimension indices that are used to chunk the big data array
        into individual Textures on the GPU
        """
        return self._zdim_indices

    def _fix_data(self, data):
        if data.ndim not in (3, 4):
            raise ValueError(
                "Volume Image data must be 3D with or without an RGB(A) dimension, i.e. "
                "it must be of shape [z, rows, cols], [z, rows, cols, 3] or [z, rows, cols, 4]"
            )

        # let's just cast to float32 always
        return data.astype(np.float32)

    def __iter__(self):
        self._iter = product(
            enumerate(self.zdim_indices),
            enumerate(self.row_indices),
            enumerate(self.col_indices),
        )

        return self

    def __next__(
        self,
    ) -> tuple[pygfx.Texture, tuple[int, int, int], tuple[slice, slice, slice]]:
        """
        Iterate through each Texture within the texture array

        Returns
        -------
        Texture, tuple[int, int], tuple[slice, slice]
            | Texture: pygfx.Texture
            | tuple[int, int]: chunk index, i.e corresponding index of ``self.buffer`` array
            | tuple[slice, slice]: data slice of big array in this chunk and Texture
        """
        # chunk indices
        (
            (chunk_z, data_z_start),
            (chunk_row, data_row_start),
            (chunk_col, data_col_start),
        ) = next(self._iter)

        # indices for to self.buffer for this chunk
        chunk_index = (chunk_z, chunk_row, chunk_col)

        # stop indices of big data array for this chunk
        z_stop = min(self.value.shape[0], data_z_start + self._texture_size_limit)
        row_stop = min(self.value.shape[1], data_row_start + self._texture_size_limit)
        col_stop = min(self.value.shape[2], data_col_start + self._texture_size_limit)

        # zdim, row and column slices that slice the data for this chunk from the big data array
        data_slice = (
            slice(data_z_start, z_stop),
            slice(data_row_start, row_stop),
            slice(data_col_start, col_stop),
        )

        # texture for this chunk
        texture = self.buffer[chunk_index]

        return texture, chunk_index, data_slice

    def __getitem__(self, item):
        return self.value[item]

    @block_reentrance
    def __setitem__(self, key, value):
        self.value[key] = value

        for texture in self.buffer.ravel():
            texture.update_range((0, 0, 0), texture.size)

        event = GraphicFeatureEvent("data", info={"key": key, "value": value})
        self._call_event_handlers(event)

    def __len__(self):
        return self.buffer.size


def create_volume_material_kwargs(graphic, mode: str):
    kwargs = {
        "clim": (graphic.vmin, graphic.vmax),
        "map": graphic._texture_map,
        "interpolation": graphic.interpolation,
        "pick_write": True,
    }

    if mode == "iso":
        more_kwargs = {
            attr: getattr(graphic, attr)
            for attr in [
                "threshold",
                "step_size",
                "substep_size",
                "emissive",
                "shininess",
            ]
        }

    elif mode == "slice":
        more_kwargs = {"plane": graphic.plane}
    else:
        more_kwargs = {}

    kwargs.update(more_kwargs)
    return kwargs


class VolumeRenderMode(GraphicFeature):
    """Volume rendering mode, controls world object material"""

    event_info_spec = [
        {
            "dict key": "value",
            "type": "str",
            "description": "volume rendering mode that has been set",
        },
    ]

    def __init__(self, value: str):
        self._validate(value)
        self._value = value
        super().__init__()

    @property
    def value(self) -> str:
        return self._value

    def _validate(self, value):
        if value not in VOLUME_RENDER_MODES.keys():
            raise ValueError(
                f"Given render mode: {value} is invalid. Valid render modes are: {VOLUME_RENDER_MODES.keys()}"
            )

    @block_reentrance
    def set_value(self, graphic, value: str):
        self._validate(value)

        VolumeMaterialCls = VOLUME_RENDER_MODES[value]

        kwargs = create_volume_material_kwargs(graphic, mode=value)

        new_material = VolumeMaterialCls(**kwargs)
        # since the world object is a group
        for volume_tile in graphic.world_object.children:
            volume_tile.material = new_material

        # so we have one place to reference it
        graphic._material = new_material
        self._value = value

        event = GraphicFeatureEvent(type="mode", info={"value": value})
        self._call_event_handlers(event)


class VolumeIsoThreshold(GraphicFeature):
    """Isosurface threshold"""

    event_info_spec = [
        {
            "dict key": "value",
            "type": "float",
            "description": "new isosurface threshold",
        },
    ]

    def __init__(self, value: float):
        self._value = value
        super().__init__()

    @property
    def value(self) -> float:
        return self._value

    @block_reentrance
    def set_value(self, graphic, value: float):
        graphic._material.threshold = value
        self._value = graphic._material.threshold

        event = GraphicFeatureEvent(type="threshold", info={"value": value})
        self._call_event_handlers(event)


class VolumeIsoStepSize(GraphicFeature):
    """Isosurface step_size"""

    event_info_spec = [
        {
            "dict key": "value",
            "type": "float",
            "description": "new isosurface step_size",
        },
    ]

    def __init__(self, value: float):
        self._value = value
        super().__init__()

    @property
    def value(self) -> float:
        return self._value

    @block_reentrance
    def set_value(self, graphic, value: float):
        graphic._material.step_size = value
        self._value = graphic._material.step_size

        event = GraphicFeatureEvent(type="step_size", info={"value": value})
        self._call_event_handlers(event)


class VolumeIsoSubStepSize(GraphicFeature):
    """Isosurface substep_size"""

    event_info_spec = [
        {
            "dict key": "value",
            "type": "float",
            "description": "new isosurface step_size",
        },
    ]

    def __init__(self, value: float):
        self._value = value
        super().__init__()

    @property
    def value(self) -> float:
        return self._value

    @block_reentrance
    def set_value(self, graphic, value: float):
        graphic._material.substep_size = value
        self._value = graphic._material.substep_size

        event = GraphicFeatureEvent(type="step_size", info={"value": value})
        self._call_event_handlers(event)


class VolumeIsoEmissive(GraphicFeature):
    """Isosurface emissive color"""

    event_info_spec = [
        {
            "dict key": "value",
            "type": "pygfx.Color",
            "description": "new isosurface emissive color",
        },
    ]

    def __init__(self, value: pygfx.Color | str | tuple | np.ndarray):
        self._value = pygfx.Color(value)
        super().__init__()

    @property
    def value(self) -> pygfx.Color:
        return self._value

    @block_reentrance
    def set_value(self, graphic, value: pygfx.Color | str | tuple | np.ndarray):
        graphic._material.emissive = value
        self._value = graphic._material.emissive

        event = GraphicFeatureEvent(type="emissive", info={"value": value})
        self._call_event_handlers(event)


class VolumeIsoShininess(GraphicFeature):
    """Isosurface shininess"""

    event_info_spec = [
        {
            "dict key": "value",
            "type": "int",
            "description": "new isosurface shininess",
        },
    ]

    def __init__(self, value: int):
        self._value = value
        super().__init__()

    @property
    def value(self) -> int:
        return self._value

    @block_reentrance
    def set_value(self, graphic, value: float):
        graphic._material.shininess = value
        self._value = graphic._material.shininess

        event = GraphicFeatureEvent(type="shininess", info={"value": value})
        self._call_event_handlers(event)


class VolumeSlicePlane(GraphicFeature):
    """Volume plane"""

    event_info_spec = [
        {
            "dict key": "value",
            "type": "tuple[float, float, float, float]",
            "description": "new plane slice",
        },
    ]

    def __init__(self, value: tuple[float, float, float, float]):
        self._value = value
        super().__init__()

    @property
    def value(self) -> tuple[float, float, float, float]:
        return self._value

    @block_reentrance
    def set_value(self, graphic, value: tuple[float, float, float, float]):
        graphic._material.plane = value
        self._value = graphic._material.plane

        event = GraphicFeatureEvent(type="plane", info={"value": value})
        self._call_event_handlers(event)



import sys
import importlib
from pathlib import Path

import wgpu


# --- Prepare


# Ultimately, we let wgpu-py decide, but we can prime things a bit to create our
# own preferred order, by importing a Qt lib. But we only do this if no GUI has
# been imported yet.

# Qt libs that we will try to import
qt_libs = ["PySide6", "PyQt6", "PySide2", "PyQt5"]

# Other known libs that, if imported, we should probably not try to force qt
other_libs = ["glfw", "wx", "ipykernel"]

already_imported = [name for name in (qt_libs + other_libs) if name in sys.modules]
if not already_imported:
    for name in qt_libs:
        try:
            importlib.import_module(name)
        except Exception:
            pass
        else:
            break


# --- Triage


# Let wgpu do the auto gui selection
from rendercanvas import BaseRenderCanvas
from rendercanvas.auto import RenderCanvas, loop

# Get the name of the backend ('qt', 'glfw', 'jupyter')
GUI_BACKEND = RenderCanvas.__module__.split(".")[-1]
IS_JUPYTER = GUI_BACKEND == "jupyter"


# --- Some backend-specific preparations


def _notebook_print_banner():
    from ipywidgets import Image
    from IPython.display import display, HTML

    logo_path = Path(__file__).parent.parent.joinpath(
        "assets", "fastplotlib_face_logo.png"
    )

    with open(logo_path, "rb") as f:
        logo_data = f.read()

    # get small logo image
    image = Image(value=logo_data, format="png", width=300, height=55)

    # get adapters and info
    adapters = [a for a in wgpu.gpu.enumerate_adapters_sync()]
    adapters_info = [a.info for a in adapters]

    default_adapter_info = wgpu.gpu.request_adapter_sync().info
    default_ix = adapters_info.index(default_adapter_info)

    if len(adapters) < 1:
        return

    # start HTML table
    table_str = (
        "<b>Available devices:</b>"
        "<table>"
        "<tr>"
        "<th>Valid</th>"
        "<th>Device</th>"
        "<th>Type</th>"
        "<th>Backend</th>"
        "<th>Driver</th>"
        "</tr>"
    )

    # parse each adapter that WGPU found
    for ix, adapter in enumerate(adapters_info):
        atype = adapter["adapter_type"]
        backend = adapter["backend_type"]
        driver = adapter["description"]
        device = adapter["device"]

        if atype in ("DiscreteGPU", "IntegratedGPU") and backend != "OpenGL":
            charactor = chr(0x2705)  # green checkmark
            tooltip = "This adapter can be used with fastplotlib"
        elif backend == "OpenGL":
            charactor = chr(0x0000274C)  # red x
            tooltip = "This adapter cannot be used with fastplotlib"
        elif device.startswith("llvmpipe") or atype == "CPU":
            charactor = f"{chr(0x00002757)} limited"  # red !
            tooltip = "CPU rendering support is limited and mainly for testing purposes"
        else:
            charactor = f"{chr(0x00002757)} unknown"  # red !
            tooltip = "Unknown adapter type and backend"

        if ix == default_ix:
            default = " (default) "
        else:
            default = ""

        # add row to HTML table
        table_str += f'<tr title="{tooltip}">'
        # add each element to this row
        for s in [f"{charactor}{default}", device, atype, backend, driver]:
            table_str += f"<td>{s}</td>"
        table_str += "</tr>"

    table_str += "</table>"

    # display logo and adapters table
    display(image)
    display(HTML(table_str))


if GUI_BACKEND == "jupyter":
    _notebook_print_banner()

elif GUI_BACKEND == "qt":
    from rendercanvas.qt import libname

    # Import submodules of PySide6/PyQt6/PySid2/PyQt5
    # For the way that fpl uses Qt, the supported Qt libs seems compatible enough.
    # If necessary we can do some qtpy-like monkey-patching here.
    QtCore = importlib.import_module(".QtCore", libname)
    QtGui = importlib.import_module(".QtGui", libname)
    QtWidgets = importlib.import_module(".QtWidgets", libname)



from typing import Literal, Union

import numpy as np

import pygfx
from rendercanvas import BaseRenderCanvas

from ..graphics import TextGraphic
from ._utils import create_camera, create_controller
from ._plot_area import PlotArea
from ._frame import Frame
from ..graphics._axes import Axes


class Subplot(PlotArea):
    def __init__(
        self,
        parent: Union["Figure"],
        camera: Literal["2d", "3d"] | pygfx.PerspectiveCamera,
        controller: pygfx.Controller | str,
        canvas: BaseRenderCanvas | pygfx.Texture,
        rect: np.ndarray = None,
        extent: np.ndarray = None,
        resizeable: bool = True,
        renderer: pygfx.WgpuRenderer = None,
        name: str = None,
    ):
        """
        Subplot class.

        .. important::
            ``Subplot`` is not meant to be constructed directly, it only exists as part of a ``Figure``

        Parameters
        ----------
        parent: 'Figure' | None
            parent Figure instance

        camera: str or pygfx.PerspectiveCamera, default '2d'
            indicates the FOV for the camera, '2d' sets ``fov = 0``, '3d' sets ``fov = 50``.
            ``fov`` can be changed at any time.

        controller: str or pygfx.Controller, optional
            | if ``None``, uses a PanZoomController for "2d" camera or FlyController for "3d" camera.
            | if ``str``, must be one of: `"panzoom", "fly", "trackball", or "orbit"`.
            | also accepts a pygfx.Controller instance

        canvas: BaseRenderCanvas, or a pygfx.Texture
            Provides surface on which a scene will be rendered.

        renderer: WgpuRenderer
            object used to render scenes using wgpu

        name: str, optional
            name of the subplot, will appear as ``TextGraphic`` above the subplot

        """

        camera = create_camera(camera)

        controller = create_controller(controller_type=controller, camera=camera)

        self._docks = dict()

        if "Imgui" in parent.__class__.__name__:
            toolbar_visible = True
        else:
            toolbar_visible = False

        super().__init__(
            parent=parent,
            camera=camera,
            controller=controller,
            scene=pygfx.Scene(),
            canvas=canvas,
            renderer=renderer,
            name=name,
        )

        for pos in ["left", "top", "right", "bottom"]:
            dv = Dock(self, size=0)
            dv.name = pos
            self.docks[pos] = dv
            self.children.append(dv)

        self._axes = Axes(self)
        self.scene.add(self.axes.world_object)

        self._frame = Frame(
            viewport=self.viewport,
            rect=rect,
            extent=extent,
            resizeable=resizeable,
            title=name,
            docks=self.docks,
            toolbar_visible=toolbar_visible,
            canvas_rect=parent.get_pygfx_render_area(),
        )

    @property
    def axes(self) -> Axes:
        """Axes object"""
        return self._axes

    @property
    def name(self) -> str:
        """Subplot name"""
        return self._name

    @name.setter
    def name(self, name: str):
        if name is None:
            self._name = None
            return

        for subplot in self.get_figure(self):
            if (subplot is self) or (subplot is None):
                continue
            if subplot.name == name:
                raise ValueError("subplot names must be unique")

        self._name = name

    @property
    def docks(self) -> dict:
        """
        The docks of this plot area. Each ``dock`` is basically just a PlotArea too.

        The docks are: ["left", "top", "right", "bottom"]

        Returns
        -------
        Dict[str, Dock]
            {dock_name: Dock}

        """
        return self._docks

    @property
    def toolbar(self) -> bool:
        """show/hide toolbar"""
        return self.frame.toolbar_visible

    @toolbar.setter
    def toolbar(self, visible: bool):
        self.frame.toolbar_visible = visible
        self.frame.reset_viewport()

    def _render(self):
        self.axes.update_using_camera()
        super()._render()

    @property
    def title(self) -> TextGraphic:
        """subplot title"""
        return self._frame.title_graphic

    @title.setter
    def title(self, text: str):
        text = str(text)
        self.title.text = text

    @property
    def frame(self) -> Frame:
        """Frame that the subplot lives in"""
        return self._frame


class Dock(PlotArea):
    def __init__(
        self,
        parent: Subplot,
        size: int,
    ):
        self._size = size

        super().__init__(
            parent=parent,
            camera=pygfx.OrthographicCamera(),
            controller=pygfx.PanZoomController(),
            scene=pygfx.Scene(),
            canvas=parent.canvas,
            renderer=parent.renderer,
        )

    @property
    def size(self) -> int:
        """Get or set the size of this dock"""
        return self._size

    @size.setter
    def size(self, s: int):
        self._size = s
        self.get_figure()._fpl_reset_layout()

    def _render(self):
        if self.size == 0:
            return

        super()._render()



import numpy as np

import pygfx

from ._base import GraphicFeature, GraphicFeatureEvent, block_reentrance


class TextData(GraphicFeature):
    event_info_spec = [
        {
            "dict key": "value",
            "type": "str",
            "description": "new text data",
        },
    ]

    def __init__(self, value: str):
        self._value = value
        super().__init__()

    @property
    def value(self) -> str:
        return self._value

    @block_reentrance
    def set_value(self, graphic, value: str):
        graphic.world_object.set_text(value)
        self._value = value

        event = GraphicFeatureEvent(type="text", info={"value": value})
        self._call_event_handlers(event)


class FontSize(GraphicFeature):
    event_info_spec = [
        {
            "dict key": "value",
            "type": "float | int",
            "description": "new font size",
        },
    ]

    def __init__(self, value: float | int):
        self._value = value
        super().__init__()

    @property
    def value(self) -> float | int:
        return self._value

    @block_reentrance
    def set_value(self, graphic, value: float | int):
        graphic.world_object.font_size = value
        self._value = graphic.world_object.font_size

        event = GraphicFeatureEvent(type="font_size", info={"value": value})
        self._call_event_handlers(event)


class TextFaceColor(GraphicFeature):
    event_info_spec = [
        {
            "dict key": "value",
            "type": "str | np.ndarray",
            "description": "new text color",
        },
    ]

    def __init__(self, value: str | np.ndarray | list[float] | tuple[float]):
        self._value = pygfx.Color(value)
        super().__init__()

    @property
    def value(self) -> pygfx.Color:
        return self._value

    @block_reentrance
    def set_value(self, graphic, value: str | np.ndarray | list[float] | tuple[float]):
        value = pygfx.Color(value)
        graphic.world_object.material.color = value
        self._value = graphic.world_object.material.color

        event = GraphicFeatureEvent(type="face_color", info={"value": value})
        self._call_event_handlers(event)


class TextOutlineColor(GraphicFeature):
    event_info_spec = [
        {
            "dict key": "value",
            "type": "str | np.ndarray",
            "description": "new outline color",
        },
    ]

    def __init__(self, value: str | np.ndarray | list[float] | tuple[float]):
        self._value = pygfx.Color(value)
        super().__init__()

    @property
    def value(self) -> pygfx.Color:
        return self._value

    @block_reentrance
    def set_value(self, graphic, value: str | np.ndarray | list[float] | tuple[float]):
        value = pygfx.Color(value)
        graphic.world_object.material.outline_color = value
        self._value = graphic.world_object.material.outline_color

        event = GraphicFeatureEvent(type="outline_color", info={"value": value})
        self._call_event_handlers(event)


class TextOutlineThickness(GraphicFeature):
    event_info_spec = [
        {
            "dict key": "value",
            "type": "float",
            "description": "new text outline thickness",
        },
    ]

    def __init__(self, value: float):
        self._value = value
        super().__init__()

    @property
    def value(self) -> float:
        return self._value

    @block_reentrance
    def set_value(self, graphic, value: float):
        graphic.world_object.material.outline_thickness = value
        self._value = graphic.world_object.material.outline_thickness

        event = GraphicFeatureEvent(type="outline_thickness", info={"value": value})
        self._call_event_handlers(event)



import os
from time import perf_counter

from imgui_bundle import imgui, icons_fontawesome_6 as fa

from ...ui import EdgeWindow


class ImageWidgetSliders(EdgeWindow):
    def __init__(self, figure, size, location, title, image_widget):
        super().__init__(figure=figure, size=size, location=location, title=title)
        self._image_widget = image_widget

        # whether or not a dimension is in play mode
        self._playing: dict[str, bool] = {"t": False, "z": False}

        # approximate framerate for playing
        self._fps: dict[str, int] = {"t": 20, "z": 20}
        # framerate converted to frame time
        self._frame_time: dict[str, float] = {"t": 1 / 20, "z": 1 / 20}

        # last timepoint that a frame was displayed from a given dimension
        self._last_frame_time: dict[str, float] = {"t": 0, "z": 0}

        self._loop = False

        if "RTD_BUILD" in os.environ.keys():
            if os.environ["RTD_BUILD"] == "1":
                self._playing["t"] = True
                self._loop = True

    def set_index(self, dim: str, index: int):
        """set the current_index of the ImageWidget"""

        # make sure the max index for this dim is not exceeded
        max_index = self._image_widget._dims_max_bounds[dim] - 1
        if index > max_index:
            if self._loop:
                # loop back to index zero if looping is enabled
                index = 0
            else:
                # if looping not enabled, stop playing this dimension
                self._playing[dim] = False
                return

        # set current_index
        self._image_widget.current_index = {dim: min(index, max_index)}

    def update(self):
        """called on every render cycle to update the GUI elements"""

        # store the new index of the image widget ("t" and "z")
        new_index = dict()

        # flag if the index changed
        flag_index_changed = False

        # reset vmin-vmax using full orig data
        if imgui.button(label=fa.ICON_FA_CIRCLE_HALF_STROKE + fa.ICON_FA_FILM):
            self._image_widget.reset_vmin_vmax()
        if imgui.is_item_hovered(0):
            imgui.set_tooltip("reset contrast limits using full movie/stack")

        # reset vmin-vmax using currently displayed ImageGraphic data
        imgui.same_line()
        if imgui.button(label=fa.ICON_FA_CIRCLE_HALF_STROKE):
            self._image_widget.reset_vmin_vmax_frame()
        if imgui.is_item_hovered(0):
            imgui.set_tooltip("reset contrast limits using current frame")

        # time now
        now = perf_counter()

        # buttons and slider UI elements for each dim
        for dim in self._image_widget.slider_dims:
            imgui.push_id(f"{self._id_counter}_{dim}")

            if self._playing[dim]:
                # show pause button if playing
                if imgui.button(label=fa.ICON_FA_PAUSE):
                    # if pause button clicked, then set playing to false
                    self._playing[dim] = False

                # if in play mode and enough time has elapsed w.r.t. the desired framerate, increment the index
                if now - self._last_frame_time[dim] >= self._frame_time[dim]:
                    self.set_index(dim, self._image_widget.current_index[dim] + 1)
                    self._last_frame_time[dim] = now

            else:
                # we are not playing, so display play button
                if imgui.button(label=fa.ICON_FA_PLAY):
                    # if play button is clicked, set last frame time to 0 so that index increments on next render
                    self._last_frame_time[dim] = 0
                    # set playing to True since play button was clicked
                    self._playing[dim] = True

            imgui.same_line()
            # step back one frame button
            if imgui.button(label=fa.ICON_FA_BACKWARD_STEP) and not self._playing[dim]:
                self.set_index(dim, self._image_widget.current_index[dim] - 1)

            imgui.same_line()
            # step forward one frame button
            if imgui.button(label=fa.ICON_FA_FORWARD_STEP) and not self._playing[dim]:
                self.set_index(dim, self._image_widget.current_index[dim] + 1)

            imgui.same_line()
            # stop button
            if imgui.button(label=fa.ICON_FA_STOP):
                self._playing[dim] = False
                self._last_frame_time[dim] = 0
                self.set_index(dim, 0)

            imgui.same_line()
            # loop checkbox
            _, self._loop = imgui.checkbox(label=fa.ICON_FA_ROTATE, v=self._loop)
            if imgui.is_item_hovered(0):
                imgui.set_tooltip("loop playback")

            imgui.same_line()
            imgui.text("framerate :")
            imgui.same_line()
            imgui.set_next_item_width(100)
            # framerate int entry
            fps_changed, value = imgui.input_int(
                label="fps", v=self._fps[dim], step_fast=5
            )
            if imgui.is_item_hovered(0):
                imgui.set_tooltip(
                    "framerate is approximate and less reliable as it approaches your monitor refresh rate"
                )
            if fps_changed:
                if value < 1:
                    value = 1
                if value > 50:
                    value = 50
                self._fps[dim] = value
                self._frame_time[dim] = 1 / value

            val = self._image_widget.current_index[dim]
            vmax = self._image_widget._dims_max_bounds[dim] - 1

            imgui.text(f"{dim}: ")
            imgui.same_line()
            # so that slider occupies full width
            imgui.set_next_item_width(self.width * 0.85)

            if "Jupyter" in self._image_widget.figure.canvas.__class__.__name__:
                # until https://github.com/pygfx/wgpu-py/issues/530
                flags = imgui.SliderFlags_.no_input
            else:
                # clamps to min, max if user inputs value outside these bounds
                flags = imgui.SliderFlags_.always_clamp

            # slider for this dimension
            changed, index = imgui.slider_int(
                f"{dim}", v=val, v_min=0, v_max=vmax, flags=flags
            )

            new_index[dim] = index

            # if the slider value changed for this dimension
            flag_index_changed |= changed

            imgui.pop_id()

        if flag_index_changed:
            # if any slider dim changed set the new index of the image widget
            self._image_widget.current_index = new_index

        self.size = int(imgui.get_window_height())



import pygfx
import numpy as np

from ._base import to_gpu_supported_dtype
from ...utils import make_pygfx_colors


def parse_colors(
    colors: str | np.ndarray | list[str] | tuple[str], n_colors: int | None
):
    """

    Parameters
    ----------
    colors
    n_colors

    Returns
    -------

    """

    # if provided as a numpy array of str
    if isinstance(colors, np.ndarray):
        if colors.dtype.kind in ["U", "S"]:
            colors = colors.tolist()
    # if the color is provided as a numpy array
    if isinstance(colors, np.ndarray):
        if colors.shape == (3,):  # single RGB array
            data = np.repeat(np.array([colors]), n_colors, axis=0)
        elif colors.shape == (4,):  # single RGBA array
            data = np.repeat(np.array([colors]), n_colors, axis=0)
        # else assume it's already a stack of RGBA arrays, keep this directly as the data
        elif colors.ndim == 2:
            if not (colors.shape[1] in (3, 4) and colors.shape[0] == n_colors):
                raise ValueError(
                    "Valid array color arguments must be a single RGBA array or a stack of "
                    "RGB or RGBA arrays for each datapoint in the shape [n_datapoints, 3] or [n_datapoints, 4]"
                )
            data = colors
        else:
            raise ValueError(
                "Valid array color arguments must be a single RGB(A) array or a stack of "
                "RGB(A) arrays for each datapoint in the shape [n_datapoints, 3] or [n_datapoints, 4]"
            )

    # if the color is provided as list or tuple
    elif isinstance(colors, (list, tuple)):
        # if iterable of str
        if all([isinstance(val, str) for val in colors]):
            if not len(colors) == n_colors:
                raise ValueError(
                    f"Valid iterable color arguments must be a `tuple` or `list` of `str` "
                    f"where the length of the iterable is the same as the number of datapoints."
                )

            data = np.vstack([np.array(pygfx.Color(c)) for c in colors])

        # if it's a single RGB/RGBA array as a tuple/list
        elif len(colors) in (3, 4):
            c = pygfx.Color(colors)
            data = np.repeat(np.array([c]), n_colors, axis=0)

        else:
            raise ValueError(
                f"Valid iterable color arguments must be a `tuple` or `list` representing RGBA values or "
                f"an iterable of `str` with the same length as the number of datapoints."
            )
    elif isinstance(colors, str):
        if colors == "random":
            data = np.random.rand(n_colors, 3)
        else:
            data = make_pygfx_colors(colors, n_colors)
    else:
        # assume it's a single color, use pygfx.Color to parse it
        data = make_pygfx_colors(colors, n_colors)

    return to_gpu_supported_dtype(data)



from enum import IntEnum


class RenderQueue(IntEnum):
    # Defaults by PyGfx
    background = 1000
    opaque = 2000
    opaque_with_discard = 2400
    auto = 2600
    transparent = 3000
    overlay = 4000
    # For axes and selectors we use a higher render_queue, so they get rendered later than
    # the graphics. Axes (rulers) have depth_compare '<=' and selectors don't compare depth.
    axes = 3400  # still in 'object' group
    selector = 3600  # considered in 'overlay' group



from ._positions_graphics import (
    VertexColors,
    UniformColor,
    UniformSize,
    SizeSpace,
    Thickness,
    VertexPositions,
    PointsSizesFeature,
    VertexCmap,
)
from ._image import (
    TextureArray,
    ImageCmap,
    ImageVmin,
    ImageVmax,
    ImageInterpolation,
    ImageCmapInterpolation,
)
from ._volume import (
    TextureArrayVolume,
    VolumeRenderMode,
    VolumeIsoThreshold,
    VolumeIsoStepSize,
    VolumeIsoSubStepSize,
    VolumeIsoEmissive,
    VolumeIsoShininess,
    VolumeSlicePlane,
    VOLUME_RENDER_MODES,
    create_volume_material_kwargs,
)
from ._base import (
    GraphicFeature,
    BufferManager,
    GraphicFeatureEvent,
    to_gpu_supported_dtype,
)

from ._text import (
    TextData,
    FontSize,
    TextFaceColor,
    TextOutlineColor,
    TextOutlineThickness,
)

from ._selection_features import (
    LinearSelectionFeature,
    LinearRegionSelectionFeature,
    RectangleSelectionFeature,
)
from ._common import Name, Offset, Rotation, Alpha, AlphaMode, Visible, Deleted


__all__ = [
    "VertexColors",
    "UniformColor",
    "UniformSize",
    "SizeSpace",
    "Thickness",
    "VertexPositions",
    "PointsSizesFeature",
    "VertexCmap",
    "TextureArray",
    "ImageCmap",
    "ImageVmin",
    "ImageVmax",
    "ImageInterpolation",
    "ImageCmapInterpolation",
    "TextureArrayVolume",
    "VolumeRenderMode",
    "VolumeIsoThreshold",
    "VolumeIsoStepSize",
    "VolumeIsoSubStepSize",
    "VolumeIsoEmissive",
    "VolumeIsoShininess",
    "VolumeSlicePlane",
    "TextData",
    "FontSize",
    "TextFaceColor",
    "TextOutlineColor",
    "TextOutlineThickness",
    "LinearSelectionFeature",
    "LinearRegionSelectionFeature",
    "RectangleSelectionFeature",
    "Name",
    "Offset",
    "Rotation",
    "Alpha",
    "AlphaMode",
    "Visible",
    "Deleted",
    "GraphicFeatureEvent",
]



import numpy as np

import pygfx
from pylinalg import quat_from_vecs, vec_transform_quat

from ..utils.enums import RenderQueue


GRID_PLANES = ["xy", "xz", "yz"]

CANONICAL_BAIS = np.array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]])


# very thin subclass that just adds GridMaterial properties to this world object for easier user control
class Grid(pygfx.Grid):
    @property
    def major_step(self) -> tuple[float, float]:
        """The step distance between the major grid lines."""
        return self.material.major_step

    @major_step.setter
    def major_step(self, step: tuple[float, float]):
        self.material.major_step = step

    @property
    def minor_step(self) -> tuple[float, float]:
        """The step distance between the minor grid lines."""
        return self.material.minor_step

    @minor_step.setter
    def minor_step(self, step: tuple[float, float]):
        self.material.minor_step = step

    @property
    def axis_thickness(self) -> float:
        """The thickness of the axis lines."""
        return self.material.axis_thickness

    @axis_thickness.setter
    def axis_thickness(self, thickness: float):
        self.material.axis_thickness = thickness

    @property
    def major_thickness(self) -> float:
        """The thickness of the major grid lines."""
        return self.material.major_thickness

    @major_thickness.setter
    def major_thickness(self, thickness: float):
        self.material.major_thickness = thickness

    @property
    def minor_thickness(self) -> float:
        """The thickness of the minor grid lines."""
        return self.material.minor_thickness

    @minor_thickness.setter
    def minor_thickness(self, thickness: float):
        self.material.minor_thickness = thickness

    @property
    def thickness_space(self) -> str:
        """The coordinate space in which the thicknesses are expressed.

        See :obj:`pygfx.utils.enums.CoordSpace`:
        """
        return self.material.thickness_space

    @thickness_space.setter
    def thickness_space(self, value: str):
        self.material.thickness_space = value

    @property
    def axis_color(self) -> str:
        """The color of the axis lines."""
        return self.material.axis_color

    @axis_color.setter
    def axis_color(self, color: str):
        self.material.axis_color = color

    @property
    def major_color(self) -> str:
        """The color of the major grid lines."""
        return self.material.major_color

    @major_color.setter
    def major_color(self, color: str):
        self.material.major_color = color

    @property
    def minor_color(self) -> str:
        """The color of the minor grid lines."""
        return self.material.minor_color

    @minor_color.setter
    def minor_color(self, color: str):
        self.material.minor_color = color

    @property
    def infinite(self) -> bool:
        """Whether the grid is infinite.

        If not infinite, the grid is 1x1 in world space, scaled, rotated, and
        positioned with the object's transform.

        (Infinite grids are not actually infinite. Rather they move along with
        the camera, and are sized based on the distance between the camera and
        the grid.)
        """
        return self.material.infinite

    @infinite.setter
    def infinite(self, value: str):
        self.material.infinite = value


class Grids(pygfx.Group):
    """Just a class to make accessing the grids easier"""

    def __init__(self, *, xy, xz, yz):
        super().__init__()

        self._xy = xy
        self._xz = xz
        self._yz = yz

        self.add(xy, xz, yz)

    @property
    def xy(self) -> Grid:
        """xy grid"""
        return self._xy

    @property
    def xz(self) -> Grid:
        """xz grid"""
        return self._xz

    @property
    def yz(self) -> Grid:
        """yz grid"""
        return self._yz


class Axes:
    def __init__(
        self,
        plot_area,
        intersection: tuple[int, int, int] | None = None,
        x_kwargs: dict = None,
        y_kwargs: dict = None,
        z_kwargs: dict = None,
        grids: bool = True,
        grid_kwargs: dict = None,
        auto_grid: bool = True,
        offset: np.ndarray = np.array([0.0, 0.0, 0.0]),
        basis: np.ndarray = np.array(
            [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]
        ),
    ):
        self._plot_area = plot_area

        x_kwargs = x_kwargs or {}
        y_kwargs = y_kwargs or {}
        z_kwargs = z_kwargs or {}

        generic_kwargs = dict(
            tick_size=8.0,
            line_width=2.0,
            tick_marker="tick",  # 'tick' for both-sides, 'tick_left' or 'tick_right' for one-sided
            color="#fff",
        )

        x_kwargs = dict(
            tick_side="right",
            **generic_kwargs,
            **x_kwargs,
        )

        y_kwargs = dict(
            tick_side="left",
            **generic_kwargs,
            **y_kwargs,
        )

        z_kwargs = dict(
            tick_side="left",
            **generic_kwargs,
            **z_kwargs,
        )

        # create ruler for each dim
        self._x = pygfx.Ruler(
            alpha_mode="solid", render_queue=RenderQueue.axes, **x_kwargs
        )
        self._y = pygfx.Ruler(
            alpha_mode="solid", render_queue=RenderQueue.axes, **y_kwargs
        )
        self._z = pygfx.Ruler(
            alpha_mode="solid", render_queue=RenderQueue.axes, **z_kwargs
        )

        # We render the lines and ticks as solid, but enable aa for text for prettier glyphs
        for ruler in self._x, self._y, self._z:
            ruler.line.material.depth_compare = "<="
            ruler.points.material.depth_compare = "<="
            ruler.text.material.depth_compare = "<="
            ruler.text.material.alpha_mode = "auto"
            ruler.text.material.aa = True

        self._offset = offset

        # *MUST* instantiate some start and end positions for the rulers else kernel crashes immediately
        # probably a WGPU rust panic
        self.x.start_pos = 0, 0, 0
        self.x.end_pos = 100, 0, 0
        self.x.start_value = self.x.start_pos[0] - offset[0]
        statsx = self.x.update(
            self._plot_area.camera, self._plot_area.viewport.logical_size
        )

        self.y.start_pos = 0, 0, 0
        self.y.end_pos = 0, 100, 0
        self.y.start_value = self.y.start_pos[1] - offset[1]
        statsy = self.y.update(
            self._plot_area.camera, self._plot_area.viewport.logical_size
        )

        self.z.start_pos = 0, 0, 0
        self.z.end_pos = 0, 0, 100
        self.z.start_value = self.z.start_pos[1] - offset[2]
        self.z.update(self._plot_area.camera, self._plot_area.viewport.logical_size)

        # world object for the rulers + grids
        self._world_object = pygfx.Group()

        # add rulers
        self.world_object.add(
            self.x,
            self.y,
            self.z,
        )

        # set z ruler invisible for orthographic projections for now
        if self._plot_area.camera.fov == 0:
            # TODO: allow any orientation in the future even for orthographic projections
            self.z.visible = False

        if grid_kwargs is None:
            grid_kwargs = dict()

        # The grid is a bit weird, because it makes use of transparency to fade off in the distance.
        # But w want it to write depth, so that objects that are drawn behind it are partually hidden.
        # So we set alha_mode to 'auto'. We make it draw earlier than other 'auto' objects, under the
        # assumption that most interesting stuff is in front of the grid, and artifacts behind the grid are less
        # bad than those in front. Note that fully opaque objects blend perfectly fine with the grid. Artifacts
        # should only emerge for objects that have semi-transparent fragments.
        grid_kwargs = dict(
            alpha_mode="auto",
            render_queue=RenderQueue.auto + 50,
            major_step=10,
            minor_step=1,
            thickness_space="screen",
            major_thickness=2,
            minor_thickness=0.5,
            infinite=True,
            **grid_kwargs,
        )

        if grids:
            _grids = dict()
            for plane in GRID_PLANES:
                grid = Grid(
                    geometry=None,
                    material=pygfx.GridMaterial(**grid_kwargs),
                    orientation=plane,
                    visible=False,
                )

                _grids[plane] = grid

            self._grids = Grids(**_grids)
            self.world_object.add(self._grids)

            if self._plot_area.camera.fov == 0:
                # orthographic projection, place grids far away
                self._grids.local.z = -1000

            major_step_x, major_step_y = statsx["tick_step"], statsy["tick_step"]

            self.grids.xy.material.major_step = major_step_x, major_step_y
            self.grids.xy.material.minor_step = 0.2 * major_step_x, 0.2 * major_step_y

        else:
            self._grids = False

        self._intersection = intersection
        self._auto_grid = auto_grid

        self._basis = None
        self.basis = basis

    @property
    def world_object(self) -> pygfx.WorldObject:
        return self._world_object

    @property
    def basis(self) -> np.ndarray:
        """get or set the basis, shape is [3, 3]"""
        return self._basis

    @basis.setter
    def basis(self, basis: np.ndarray):
        if basis.shape != (3, 3):
            raise ValueError

        # apply quaternion to each of x, y, z rulers
        for dim, cbasis, new_basis in zip(["x", "y", "z"], CANONICAL_BAIS, basis):
            ruler: pygfx.Ruler = getattr(self, dim)
            ruler.local.rotation = quat_from_vecs(cbasis, new_basis)

    @property
    def offset(self) -> np.ndarray:
        """offset of the axes"""
        return self._offset

    @offset.setter
    def offset(self, value: np.ndarray):
        self._offset = value

    @property
    def x(self) -> pygfx.Ruler:
        """x axis ruler"""
        return self._x

    @property
    def y(self) -> pygfx.Ruler:
        """y axis ruler"""
        return self._y

    @property
    def z(self) -> pygfx.Ruler:
        """z axis ruler"""
        return self._z

    @property
    def grids(self) -> Grids | bool:
        """grids for each plane: xy, xz, yz"""
        return self._grids

    @property
    def colors(self) -> tuple[pygfx.Color]:
        return tuple(getattr(self, dim).line.material.color for dim in ["x", "y", "z"])

    @colors.setter
    def colors(self, colors: tuple[pygfx.Color | str]):
        """get or set the colors for the x, y, and z rulers"""
        if len(colors) != 3:
            raise ValueError

        for dim, color in zip(["x", "y", "z"], colors):
            getattr(self, dim).line.material.color = color

    @property
    def auto_grid(self) -> bool:
        """auto adjust the grid on each render cycle"""
        return self._auto_grid

    @auto_grid.setter
    def auto_grid(self, value: bool):
        self._auto_grid = value

    @property
    def visible(self) -> bool:
        """set visibility of all axes elements, rulers and grids"""
        return self._world_object.visible

    @visible.setter
    def visible(self, value: bool):
        self._world_object.visible = value

    @property
    def intersection(self) -> tuple[float, float, float] | None:
        return self._intersection

    @intersection.setter
    def intersection(self, intersection: tuple[float, float, float] | None):
        """
        intersection point of [x, y, z] rulers.
        Set (0, 0, 0) for origin
        Set to `None` to follow when panning through the scene with orthographic projection
        """
        if intersection is None:
            self._intersection = None
            return

        if len(intersection) != 3:
            raise ValueError(
                "intersection must be a float of 3 elements for [x, y, z] or `None`"
            )

        self._intersection = tuple(float(v) for v in intersection)

    def update_using_bbox(self, bbox):
        """
        Update the w.r.t. the given bbox

        Parameters
        ----------
        bbox: np.ndarray
            array of shape [2, 3], [[xmin, ymin, zmin], [xmax, ymax, zmax]]

        """

        # flip axes if camera scale is flipped
        if self._plot_area.camera.local.scale_x < 0:
            bbox[0, 0], bbox[1, 0] = bbox[1, 0], bbox[0, 0]

        if self._plot_area.camera.local.scale_y < 0:
            bbox[0, 1], bbox[1, 1] = bbox[1, 1], bbox[0, 1]

        if self._plot_area.camera.local.scale_z < 0:
            bbox[0, 2], bbox[1, 2] = bbox[1, 2], bbox[0, 2]

        if self.intersection is None:
            intersection = (0, 0, 0)
        else:
            intersection = self.intersection

        self.update(bbox, intersection)

    def update_using_camera(self):
        """
        Update the axes w.r.t the current camera state

        For orthographic projections of the xy plane, it will calculate the inverse projection
        of the screen space onto world space to determine the current range of the world space
        to set the rulers and ticks

        For perspective projections it will just use the bbox of the scene to set the rulers

        """

        if not self.visible:
            return

        if self._plot_area.camera.fov == 0:
            xpos, ypos, width, height = self._plot_area.viewport.rect
            # orthographic projection, get ranges using inverse

            # get range of screen space by getting the corners
            xmin, xmax = xpos, xpos + width
            ymin, ymax = ypos + height, ypos

            # apply quaternion to account for rotation of axes
            # xmin, _, _ = vec_transform_quat(
            #     [xmin, ypos + height / 2, 0],
            #     self.x.local.rotation
            # )
            #
            # xmax, _, _ = vec_transform_quat(
            #     [xmax, ypos + height / 2, 0],
            #     self.x.local.rotation,
            # )
            #
            # _, ymin, _ = vec_transform_quat(
            #     [xpos + width / 2, ymin, 0],
            #     self.y.local.rotation
            # )
            #
            # _, ymax, _ = vec_transform_quat(
            #     [xpos + width / 2, ymax, 0],
            #     self.y.local.rotation
            # )

            min_vals = self._plot_area.map_screen_to_world((xmin, ymin))
            max_vals = self._plot_area.map_screen_to_world((xmax, ymax))

            if min_vals is None or max_vals is None:
                return

            world_xmin, world_ymin, _ = min_vals
            world_xmax, world_ymax, _ = max_vals

            world_zmin, world_zmax = 0, 0

            bbox = np.array(
                [
                    [world_xmin, world_ymin, world_zmin],
                    [world_xmax, world_ymax, world_zmax],
                ]
            )

        else:
            # set ruler start and end positions based on scene bbox
            bbox = self._plot_area._fpl_graphics_scene.get_world_bounding_box()

        if self.intersection is None:
            if self._plot_area.camera.fov == 0:
                # place the ruler close to the left and bottom edges of the viewport
                # TODO: determine this for perspective projections
                xscreen_10, yscreen_10 = xpos + (width * 0.1), ypos + (height * 0.9)
                intersection = self._plot_area.map_screen_to_world(
                    (xscreen_10, yscreen_10)
                )
            else:
                # force origin since None is not supported for Persepctive projections
                self._intersection = (0, 0, 0)
                intersection = self._intersection

        else:
            # axes intersect at the origin
            intersection = self.intersection

        self.update(bbox, intersection)

    def update(self, bbox, intersection):
        """
        Update the axes using the given bbox and ruler intersection point

        Parameters
        ----------
        bbox: np.ndarray
            array of shape [2, 3], [[xmin, ymin, zmin], [xmax, ymax, zmax]]

        intersection: float, float, float
            intersection point of the x, y, z ruler

        """

        world_xmin, world_ymin, world_zmin = bbox[0]
        world_xmax, world_ymax, world_zmax = bbox[1]
        world_x_10, world_y_10, world_z_10 = intersection

        # swap min and max for each dimension if necessary
        if self._plot_area.camera.local.scale_y < 0:
            world_ymin, world_ymax = world_ymax, world_ymin
            self.y.tick_side = "right"  # swap tick side
            self.x.tick_side = "right"
        else:
            self.y.tick_side = "left"
            self.x.tick_side = "right"

        if self._plot_area.camera.local.scale_x < 0:
            world_xmin, world_xmax = world_xmax, world_xmin
            self.x.tick_side = "left"

        self.x.start_pos = world_xmin, world_y_10, world_z_10
        self.x.end_pos = world_xmax, world_y_10, world_z_10

        self.x.start_value = self.x.start_pos[0] - self.offset[0]
        statsx = self.x.update(
            self._plot_area.camera, self._plot_area.viewport.logical_size
        )

        self.y.start_pos = world_x_10, world_ymin, world_z_10
        self.y.end_pos = world_x_10, world_ymax, world_z_10

        self.y.start_value = self.y.start_pos[1] - self.offset[1]
        statsy = self.y.update(
            self._plot_area.camera, self._plot_area.viewport.logical_size
        )

        if self._plot_area.camera.fov != 0:
            self.z.start_pos = world_x_10, world_y_10, world_zmin
            self.z.end_pos = world_x_10, world_y_10, world_zmax

            self.z.start_value = self.z.start_pos[2] - self.offset[2]
            statsz = self.z.update(
                self._plot_area.camera, self._plot_area.viewport.logical_size
            )
            major_step_z = statsz["tick_step"]

        if self.grids:
            if self.auto_grid:
                major_step_x, major_step_y = statsx["tick_step"], statsy["tick_step"]
                self.grids.xy.major_step = major_step_x, major_step_y
                self.grids.xy.minor_step = 0.2 * major_step_x, 0.2 * major_step_y

                if self._plot_area.camera.fov != 0:
                    self.grids.xz.major_step = major_step_x, major_step_z
                    self.grids.xz.minor_step = 0.2 * major_step_x, 0.2 * major_step_z

                    self.grids.yz.material.major_step = major_step_y, major_step_z
                    self.grids.yz.minor_step = 0.2 * major_step_y, 0.2 * major_step_z



from typing import *

import numpy as np
import pygfx

from ._positions_base import PositionsGraphic
from .features import (
    PointsSizesFeature,
    UniformSize,
    SizeSpace,
    VertexPositions,
    VertexColors,
    UniformColor,
    VertexCmap,
)


class ScatterGraphic(PositionsGraphic):
    _features = {
        "data": VertexPositions,
        "sizes": (PointsSizesFeature, UniformSize),
        "colors": (VertexColors, UniformColor),
        "cmap": (VertexCmap, None),
        "size_space": SizeSpace,
    }

    def __init__(
        self,
        data: Any,
        colors: str | np.ndarray | tuple[float] | list[float] | list[str] = "w",
        uniform_color: bool = False,
        cmap: str = None,
        cmap_transform: np.ndarray = None,
        isolated_buffer: bool = True,
        sizes: float | np.ndarray | Sequence[float] = 1,
        uniform_size: bool = False,
        size_space: str = "screen",
        **kwargs,
    ):
        """
        Create a Scatter Graphic, 2d or 3d

        Parameters
        ----------
        data: array-like
            Scatter data to plot, Can provide 2D, or a 3D data. 2D data must be of shape [n_points, 2].
            3D data must be of shape [n_points, 3]

        colors: str, array, tuple, list, Sequence, default "w"
            specify colors as a single human-readable string, a single RGBA array,
            or a Sequence (array, tuple, or list) of strings or RGBA arrays

        uniform_color: bool, default False
            if True, uses a uniform buffer for the scatter point colors. Useful if you need to
            save GPU VRAM when all points have the same color.

        cmap: str, optional
            apply a colormap to the scatter instead of assigning colors manually, this
            overrides any argument passed to "colors".  For supported colormaps see the
            ``cmap`` library catalogue: https://cmap-docs.readthedocs.io/en/stable/catalog/

        cmap_transform: 1D array-like or list of numerical values, optional
            if provided, these values are used to map the colors from the cmap

        isolated_buffer: bool, default True
            whether the buffers should be isolated from the user input array.
            Generally always ``True``, ``False`` is for rare advanced use if you have large arrays.

        sizes: float or iterable of float, optional, default 1.0
            sizes of the scatter points

        uniform_size: bool, default False
            if True, uses a uniform buffer for the scatter point sizes. Useful if you need to
            save GPU VRAM when all points have the same size.

        size_space: str, default "screen"
            coordinate space in which the size is expressed ("screen", "world", "model")

        kwargs
            passed to :class:`.Graphic`

        """

        super().__init__(
            data=data,
            colors=colors,
            uniform_color=uniform_color,
            cmap=cmap,
            cmap_transform=cmap_transform,
            isolated_buffer=isolated_buffer,
            size_space=size_space,
            **kwargs,
        )

        n_datapoints = self.data.value.shape[0]

        aa = kwargs.get("alpha_mode", "auto") in ("blend", "weighted_blend")

        geo_kwargs = {"positions": self._data.buffer}
        material_kwargs = dict(pick_write=True, aa=aa, depth_compare="<=")
        self._size_space = SizeSpace(size_space)

        if uniform_color:
            material_kwargs["color_mode"] = "uniform"
            material_kwargs["color"] = self.colors
        else:
            material_kwargs["color_mode"] = "vertex"
            geo_kwargs["colors"] = self.colors.buffer

        if uniform_size:
            material_kwargs["size_mode"] = "uniform"
            self._sizes = UniformSize(sizes)
            material_kwargs["size"] = self.sizes
        else:
            material_kwargs["size_mode"] = "vertex"
            self._sizes = PointsSizesFeature(sizes, n_datapoints=n_datapoints)
            geo_kwargs["sizes"] = self.sizes.buffer

        material_kwargs["size_space"] = self.size_space
        world_object = pygfx.Points(
            pygfx.Geometry(**geo_kwargs),
            material=pygfx.PointsMaterial(**material_kwargs),
        )

        self._set_world_object(world_object)

    @property
    def sizes(self) -> PointsSizesFeature | float:
        """Get or set the scatter point size(s)"""
        if isinstance(self._sizes, PointsSizesFeature):
            return self._sizes

        elif isinstance(self._sizes, UniformSize):
            return self._sizes.value

    @sizes.setter
    def sizes(self, value):
        if isinstance(self._sizes, PointsSizesFeature):
            self._sizes[:] = value

        elif isinstance(self._sizes, UniformSize):
            self._sizes.set_value(self, value)



from imgui_bundle import imgui

from ...layouts._utils import controller_types
from ...layouts._plot_area import PlotArea
from ...ui import Popup


def flip_axis(subplot: PlotArea, axis: str, flip: bool):
    camera = subplot.camera
    axis_attr = f"scale_{axis}"
    scale = getattr(camera.local, axis_attr)

    if flip and scale > 0:
        # flip is checked and axis is not already flipped
        setattr(camera.local, axis_attr, scale * -1)

    elif not flip and scale < 0:
        # flip not checked and axis is flipped
        setattr(camera.local, axis_attr, scale * -1)


class StandardRightClickMenu(Popup):
    """Right click menu that is shown on subplots"""

    def __init__(self, figure):
        super().__init__(figure=figure)

        self._last_right_click_pos = None
        self._mouse_down: bool = False

        # whether the right click menu is currently open or not
        self.is_open: bool = False

    def get_subplot(self) -> PlotArea | bool | None:
        """get the subplot that a click occurred in"""
        if self._last_right_click_pos is None:
            return False

        for subplot in self._figure:
            if subplot.viewport.is_inside(*self._last_right_click_pos):
                return subplot

        # not inside a subplot
        return False

    def cleanup(self):
        """called when the popup disappears"""
        self.is_open = False

    def update(self):
        if imgui.is_mouse_down(1) and not self._mouse_down:
            # mouse button was pressed down, store this position
            self._mouse_down = True
            self._last_right_click_pos = imgui.get_mouse_pos()

        if imgui.is_mouse_released(1) and self._mouse_down:
            self._mouse_down = False

            # open popup only if mouse was not moved between mouse_down and mouse_up events
            if self._last_right_click_pos == imgui.get_mouse_pos():
                if self.get_subplot() is not False:  # must explicitly check for False
                    # open only if right click was inside a subplot
                    imgui.open_popup(f"right-click-menu")

        # TODO: call this just once when going from open -> closed state
        if not imgui.is_popup_open("right-click-menu"):
            self.cleanup()

        if imgui.begin_popup(f"right-click-menu"):
            if self.get_subplot() is False:  # must explicitly check for False
                # for some reason it will still trigger at certain locations
                # despite open_popup() only being called when an actual
                # subplot is returned
                imgui.end_popup()
                imgui.close_current_popup()
                self.cleanup()
                return

            name = self.get_subplot().name

            if name is not None:
                # text label at the top of the menu
                imgui.text(f"subplot: {name}")
                imgui.separator()

            _, show_fps = imgui.menu_item(
                "Show fps", "", self.get_subplot().get_figure().imgui_show_fps
            )
            self.get_subplot().get_figure().imgui_show_fps = show_fps

            # autoscale, center, maintain aspect
            if imgui.menu_item(f"Autoscale", "", False)[0]:
                self.get_subplot().auto_scale()

            if imgui.menu_item(f"Center", "", False)[0]:
                self.get_subplot().center_scene()

            _, maintain_aspect = imgui.menu_item(
                "Maintain Aspect", "", self.get_subplot().camera.maintain_aspect
            )
            self.get_subplot().camera.maintain_aspect = maintain_aspect

            imgui.separator()

            # toggles to flip axes cameras
            for axis in ["x", "y", "z"]:
                scale = getattr(self.get_subplot().camera.local, f"scale_{axis}")
                changed, flip = imgui.menu_item(
                    f"Flip {axis} axis", "", bool(scale < 0)
                )

                if changed:
                    flip_axis(self.get_subplot(), axis, flip)

            imgui.separator()

            # toggles to show/hide the grid
            for plane in ["xy", "xz", "yz"]:
                grid = getattr(self.get_subplot().axes.grids, plane)
                visible = grid.visible
                changed, new_visible = imgui.menu_item(f"Grid {plane}", "", visible)

                if changed:
                    grid.visible = new_visible

            imgui.separator()

            # camera FOV
            changed, fov = imgui.slider_float(
                "FOV", v=self.get_subplot().camera.fov, v_min=0.0, v_max=180.0
            )

            imgui.separator()

            if changed:
                # FOV between 0 and 1 is numerically unstable
                if 0 < fov < 1:
                    fov = 1

                # need to update FOV via controller, if FOV is directly set
                # on the camera the controller will immediately set it back
                self.get_subplot().controller.update_fov(
                    fov - self.get_subplot().camera.fov,
                    animate=False,
                )

            imgui.separator()

            # controller options
            if imgui.begin_menu("Controller"):
                _, enabled = imgui.menu_item(
                    "Enabled", "", self.get_subplot().controller.enabled
                )

                self.get_subplot().controller.enabled = enabled

                changed, damping = imgui.slider_float(
                    "Damping",
                    v=self.get_subplot().controller.damping,
                    v_min=0.0,
                    v_max=10.0,
                )

                if changed:
                    self.get_subplot().controller.damping = damping

                imgui.separator()
                imgui.text("Controller type:")
                # switching between different controllers
                for name, controller_type_iter in controller_types.items():
                    current_type = type(self.get_subplot().controller)

                    clicked, _ = imgui.menu_item(
                        label=name,
                        shortcut="",
                        p_selected=current_type is controller_type_iter,
                    )

                    if clicked and (current_type is not controller_type_iter):
                        # menu item was clicked and the desired controller isn't the current one
                        self.get_subplot().controller = name

                imgui.end_menu()

            imgui.end_popup()



from pathlib import Path
from typing import Literal, Iterable

import numpy as np

import imgui_bundle
from imgui_bundle import imgui, icons_fontawesome_6 as fa

from wgpu.utils.imgui import ImguiRenderer, Stats
from rendercanvas import BaseRenderCanvas

import pygfx

from ._figure import Figure
from ..ui import EdgeWindow, SubplotToolbar, StandardRightClickMenu, Popup, GUI_EDGES
from ..ui import ColormapPicker


class ImguiFigure(Figure):
    def __init__(
        self,
        shape: tuple[int, int] = (1, 1),
        rects: list[tuple | np.ndarray] = None,
        extents: list[tuple | np.ndarray] = None,
        cameras: (
            Literal["2d", "3d"]
            | Iterable[Iterable[Literal["2d", "3d"]]]
            | pygfx.PerspectiveCamera
            | Iterable[Iterable[pygfx.PerspectiveCamera]]
        ) = "2d",
        controller_types: (
            Iterable[Iterable[Literal["panzoom", "fly", "trackball", "orbit"]]]
            | Iterable[Literal["panzoom", "fly", "trackball", "orbit"]]
        ) = None,
        controller_ids: (
            Literal["sync"]
            | Iterable[int]
            | Iterable[Iterable[int]]
            | Iterable[Iterable[str]]
        ) = None,
        controllers: pygfx.Controller | Iterable[Iterable[pygfx.Controller]] = None,
        canvas: str | BaseRenderCanvas | pygfx.Texture = None,
        renderer: pygfx.WgpuRenderer = None,
        canvas_kwargs: dict = None,
        size: tuple[int, int] = (500, 300),
        names: list | np.ndarray = None,
        show_tooltips: bool = False,
    ):
        self._guis: dict[str, EdgeWindow] = {k: None for k in GUI_EDGES}

        super().__init__(
            shape=shape,
            rects=rects,
            extents=extents,
            cameras=cameras,
            controller_types=controller_types,
            controller_ids=controller_ids,
            controllers=controllers,
            canvas=canvas,
            renderer=renderer,
            canvas_kwargs=canvas_kwargs,
            size=size,
            names=names,
            show_tooltips=show_tooltips,
        )

        self._imgui_renderer = ImguiRenderer(self.renderer.device, self.canvas)

        # This loads both the Roboto Font and FontAwesome 6 icons and creates and merged font
        # allowing us to use both without pushing and popping to display icons or regular text
        sans_serif_font = str(
            Path(imgui_bundle.__file__).parent.joinpath(
                "assets", "fonts", "Roboto", "Roboto-Regular.ttf"
            )
        )

        fa_6_fonts_path = str(
            Path(imgui_bundle.__file__).parent.joinpath(
                "assets", "fonts", "Font_Awesome_6_Free-Solid-900.otf"
            )
        )

        io = imgui.get_io()

        self._default_imgui_font = io.fonts.add_font_from_file_ttf(
            sans_serif_font, 14, imgui.ImFontConfig()
        )

        font_config = imgui.ImFontConfig()
        font_config.merge_mode = True

        self._default_imgui_font = io.fonts.add_font_from_file_ttf(
            fa_6_fonts_path,
            14,
            font_config,
        )

        imgui.push_font(self._default_imgui_font, self._default_imgui_font.legacy_size)

        self.imgui_renderer.set_gui(self._draw_imgui)

        self._subplot_toolbars: np.ndarray[SubplotToolbar] = np.empty(
            shape=self._subplots.size, dtype=object
        )

        for i, subplot in enumerate(self._subplots.ravel()):
            toolbar = SubplotToolbar(subplot=subplot)
            self._subplot_toolbars[i] = toolbar

        self._right_click_menu = StandardRightClickMenu(figure=self)

        self._popups: dict[str, Popup] = {}

        self.imgui_show_fps = False
        self._stats = Stats(self.renderer.device, self.canvas)

        self.register_popup(ColormapPicker)

    @property
    def default_imgui_font(self) -> imgui.ImFont:
        return self._default_imgui_font

    @property
    def guis(self) -> dict[str, EdgeWindow]:
        """GUI windows added to the Figure"""
        return self._guis

    @property
    def imgui_renderer(self) -> ImguiRenderer:
        """imgui renderer"""
        return self._imgui_renderer

    def _render(self, draw=False):
        if self.imgui_show_fps:
            with self._stats:
                super()._render(draw)
        else:
            super()._render(draw)

        self.imgui_renderer.render()

        # needs to be here else events don't get processed
        self.canvas.request_draw()

    def _draw_imgui(self) -> imgui.ImDrawData:
        # imgui.new_frame()

        for subplot, toolbar in zip(
            self._subplots.ravel(), self._subplot_toolbars.ravel()
        ):
            if not subplot.toolbar:
                # if subplot.toolbar is False
                continue
            toolbar.update()

        for gui in self.guis.values():
            if gui is not None:
                gui.draw_window()

        for popup in self._popups.values():
            popup.update()

        self._right_click_menu.update()

        # imgui.end_frame()

        # imgui.render()

        # return imgui.get_draw_data()

    def add_gui(self, gui: EdgeWindow):
        """
        Add a GUI to the Figure. GUIs can be added to the left or bottom edge.

        Parameters
        ----------
        gui: EdgeWindow
            A GUI EdgeWindow instance

        """
        if not isinstance(gui, EdgeWindow):
            raise TypeError(
                f"GUI must be of type: {EdgeWindow} you have passed a {type(gui)}"
            )

        location = gui.location

        if location not in GUI_EDGES:
            raise ValueError(
                f"GUI does not have a valid location, valid locations are: {GUI_EDGES}, you have passed: {location}"
            )

        if self.guis[location] is not None:
            raise ValueError(f"GUI already exists in the desired location: {location}")

        self.guis[location] = gui

        self._fpl_reset_layout()

    def get_pygfx_render_area(self, *args) -> tuple[int, int, int, int]:
        """
        Get rect for the portion of the canvas that the pygfx renderer draws to,
        i.e. non-imgui, part of canvas

        Returns
        -------
        tuple[int, int, int, int]
            x_pos, y_pos, width, height

        """

        width, height = self.canvas.get_logical_size()

        for edge in ["right"]:
            if self.guis[edge]:
                width -= self._guis[edge].size

        for edge in ["bottom"]:
            if self.guis[edge]:
                height -= self._guis[edge].size

        return 0, 0, max(1, width), max(1, height)

    def register_popup(self, popup: Popup.__class__):
        """
        Register a popup class. Note that this takes the class, not an instance

        Parameters
        ----------
        popup: Popup subclass

        """
        self._popups[popup.name] = popup(self)

    def open_popup(self, name: str, pos: tuple[int, int], **kwargs):
        """
        Open a registered popup

        Parameters
        ----------
        name: str
            The registered name of the popup

        pos: int, int
            x_pos, y_pos for the popup

        kwargs
            any additional kwargs to pass to the Popup's open() method

        """

        if self._popups[name].is_open:
            return

        self._popups[name].open(pos, **kwargs)



from typing import *

import numpy as np

import pygfx

from ._positions_base import PositionsGraphic
from .selectors import (
    LinearRegionSelector,
    LinearSelector,
    RectangleSelector,
    PolygonSelector,
)
from .features import (
    Thickness,
    VertexPositions,
    VertexColors,
    UniformColor,
    VertexCmap,
    SizeSpace,
)
from ..utils import quick_min_max


class LineGraphic(PositionsGraphic):
    _features = {
        "data": VertexPositions,
        "colors": (VertexColors, UniformColor),
        "cmap": (VertexCmap, None),  # none if UniformColor
        "thickness": Thickness,
        "size_space": SizeSpace,
    }

    def __init__(
        self,
        data: Any,
        thickness: float = 2.0,
        colors: str | np.ndarray | Sequence = "w",
        uniform_color: bool = False,
        cmap: str = None,
        cmap_transform: np.ndarray | Sequence = None,
        isolated_buffer: bool = True,
        size_space: str = "screen",
        **kwargs,
    ):
        """
        Create a line Graphic, 2d or 3d

        Parameters
        ----------
        data: array-like
            Line data to plot. Can provide 1D, 2D, or a 3D data.
            | If passing a 1D array, it is used to set the y-values and the x-values are generated as an integer range
            from [0, data.size]
            | 2D data must be of shape [n_points, 2]. 3D data must be of shape [n_points, 3]

        thickness: float, optional, default 2.0
            thickness of the line

        colors: str, array, or iterable, default "w"
            specify colors as a single human-readable string, a single RGBA array,
            or a Sequence (array, tuple, or list) of strings or RGBA arrays

        uniform_color: bool, default ``False``
            if True, uses a uniform buffer for the line color,
            basically saves GPU VRAM when the entire line has a single color

        cmap: str, optional
            Apply a colormap to the line instead of assigning colors manually, this
            overrides any argument passed to "colors". For supported colormaps see the
            ``cmap`` library catalogue: https://cmap-docs.readthedocs.io/en/stable/catalog/

        cmap_transform: 1D array-like of numerical values, optional
            if provided, these values are used to map the colors from the cmap

        size_space: str, default "screen"
            coordinate space in which the thickness is expressed ("screen", "world", "model")

        **kwargs
            passed to :class:`.Graphic`

        """

        super().__init__(
            data=data,
            colors=colors,
            uniform_color=uniform_color,
            cmap=cmap,
            cmap_transform=cmap_transform,
            isolated_buffer=isolated_buffer,
            size_space=size_space,
            **kwargs,
        )

        self._thickness = Thickness(thickness)

        if thickness < 1.1:
            MaterialCls = pygfx.LineThinMaterial
            aa = True
        else:
            MaterialCls = pygfx.LineMaterial

        aa = kwargs.get("alpha_mode", "auto") in ("blend", "weighted_blend")

        if uniform_color:
            geometry = pygfx.Geometry(positions=self._data.buffer)
            material = MaterialCls(
                aa=aa,
                thickness=self.thickness,
                color_mode="uniform",
                color=self.colors,
                pick_write=True,
                thickness_space=self.size_space,
                depth_compare="<=",
            )
        else:
            material = MaterialCls(
                aa=aa,
                thickness=self.thickness,
                color_mode="vertex",
                pick_write=True,
                thickness_space=self.size_space,
                depth_compare="<=",
            )
            geometry = pygfx.Geometry(
                positions=self._data.buffer, colors=self._colors.buffer
            )

        world_object: pygfx.Line = pygfx.Line(geometry=geometry, material=material)

        self._set_world_object(world_object)

    @property
    def thickness(self) -> float:
        """Get or set the line thickness"""
        return self._thickness.value

    @thickness.setter
    def thickness(self, value: float):
        self._thickness.set_value(self, value)

    def add_linear_selector(
        self, selection: float = None, axis: str = "x", **kwargs
    ) -> LinearSelector:
        """
        Adds a :class:`.LinearSelector`.

        Selectors are just ``Graphic`` objects, so you can manage, remove, or delete them from a
        plot area just like any other ``Graphic``.

        Parameters
        ----------
        selection: float, optional
            selected point on the linear selector, by default the first datapoint on the line.

        axis: str, default "x"
            axis that the selector resides on

        kwargs
            passed to :class:`.LinearSelector`

        Returns
        -------
        LinearSelector

        """

        bounds_init, limits, size, center = self._get_linear_selector_init_args(
            axis, padding=0
        )

        if selection is None:
            selection = bounds_init[0]

        selector = LinearSelector(
            selection=selection,
            limits=limits,
            axis=axis,
            parent=self,
            **kwargs,
        )

        self._plot_area.add_graphic(selector, center=False)

        return selector

    def add_linear_region_selector(
        self,
        selection: tuple[float, float] = None,
        padding: float = 0.0,
        axis: str = "x",
        **kwargs,
    ) -> LinearRegionSelector:
        """
        Add a :class:`.LinearRegionSelector`.

        Selectors are just ``Graphic`` objects, so you can manage, remove, or delete them from a
        plot area just like any other ``Graphic``.

        Parameters
        ----------
        selection: (float, float), optional
            the starting bounds of the linear region selector, computed from data if not provided

        axis: str, default "x"
            axis that the selector resides on

        padding: float, default 0.0
            Extra padding to extend the linear region selector along the orthogonal axis to make it easier to interact with.

        kwargs
            passed to ``LinearRegionSelector``

        Returns
        -------
        LinearRegionSelector
            linear selection graphic

        """

        bounds_init, limits, size, center = self._get_linear_selector_init_args(
            axis, padding
        )

        if selection is None:
            selection = bounds_init

        # create selector
        selector = LinearRegionSelector(
            selection=selection,
            limits=limits,
            size=size,
            center=center,
            axis=axis,
            parent=self,
            **kwargs,
        )

        self._plot_area.add_graphic(selector, center=False)

        # PlotArea manages this for garbage collection etc. just like all other Graphics
        # so we should only work with a proxy on the user-end
        return selector

    def add_rectangle_selector(
        self,
        selection: tuple[float, float, float, float] = None,
        **kwargs,
    ) -> RectangleSelector:
        """
        Add a :class:`.RectangleSelector`.

        Selectors are just ``Graphic`` objects, so you can manage, remove, or delete them from a
        plot area just like any other ``Graphic``.

        Parameters
        ----------
        selection: (float, float, float, float), optional
            initial (xmin, xmax, ymin, ymax) of the selection
        """
        # computes args to create selectors
        n_datapoints = self.data.value.shape[0]
        value_25p = int(n_datapoints / 4)

        # remove any nans
        data = self.data.value[~np.any(np.isnan(self.data.value), axis=1)]

        x_axis_vals = data[:, 0]
        y_axis_vals = data[:, 1]

        ymin = np.floor(y_axis_vals.min()).astype(int)
        ymax = np.ceil(y_axis_vals.max()).astype(int)

        # default selection is 25% of the image
        if selection is None:
            selection = (x_axis_vals[0], x_axis_vals[value_25p], ymin, ymax)

        # min/max limits
        limits = (x_axis_vals[0], x_axis_vals[-1], ymin * 1.5, ymax * 1.5)

        selector = RectangleSelector(
            selection=selection,
            limits=limits,
            parent=self,
            **kwargs,
        )

        self._plot_area.add_graphic(selector, center=False)

        return selector

    def add_polygon_selector(
        self,
        selection: List[tuple[float, float]] = None,
        **kwargs,
    ) -> PolygonSelector:
        """
        Add a :class:`.PolygonSelector`.

        Selectors are just ``Graphic`` objects, so you can manage, remove, or delete them from a
        plot area just like any other ``Graphic``.

        Parameters
        ----------
        selection: List of positions, optional
            Initial points for the polygon. If not given or None, you'll start drawing the selection (clicking adds points to the polygon).
        """

        # remove any nans
        data = self.data.value[~np.any(np.isnan(self.data.value), axis=1)]

        x_axis_vals = data[:, 0]
        y_axis_vals = data[:, 1]

        ymin = np.floor(y_axis_vals.min()).astype(int)
        ymax = np.ceil(y_axis_vals.max()).astype(int)

        # min/max limits
        limits = (x_axis_vals[0], x_axis_vals[-1], ymin * 1.5, ymax * 1.5)

        selector = PolygonSelector(
            selection,
            limits,
            parent=self,
            **kwargs,
        )

        self._plot_area.add_graphic(selector, center=False)

        return selector

    # TODO: this method is a bit of a mess, can refactor later
    def _get_linear_selector_init_args(
        self, axis: str, padding
    ) -> tuple[tuple[float, float], tuple[float, float], float, float]:
        # computes args to create selectors
        n_datapoints = self.data.value.shape[0]
        value_25p = int(n_datapoints / 4)

        # remove any nans
        data = self.data.value[~np.any(np.isnan(self.data.value), axis=1)]

        if axis == "x":
            # xvals
            axis_vals = data[:, 0]

            # yvals to get size and center
            magn_vals = data[:, 1]
        elif axis == "y":
            axis_vals = data[:, 1]
            magn_vals = data[:, 0]

        bounds_init = axis_vals[0], axis_vals[value_25p]
        limits = axis_vals[0], axis_vals[-1]

        # width or height of selector
        size = int(np.ptp(magn_vals) * 1.5 + padding)

        # center of selector along the other axis
        center = sum(quick_min_max(magn_vals)) / 2

        return bounds_init, limits, size, center



import math
from numbers import Real
from typing import Sequence

import numpy as np
import pygfx

from ...utils.enums import RenderQueue
from .._base import Graphic
from .._collection_base import GraphicCollection
from ..features._selection_features import LinearSelectionFeature
from ._base_selector import BaseSelector, MoveInfo


class LinearSelector(BaseSelector):
    _features = {"selection": LinearSelectionFeature}

    @property
    def parent(self) -> Graphic:
        return self._parent

    @property
    def selection(self) -> float:
        """
        x or y value of selector's current position
        """
        return self._selection.value

    @selection.setter
    def selection(self, value: int):
        graphic = self._parent

        if isinstance(graphic, GraphicCollection):
            pass

        self._selection.set_value(self, value)

    @property
    def limits(self) -> tuple[float, float]:
        return self._limits

    @limits.setter
    def limits(self, values: tuple[float, float]):
        # check that `values` is an iterable of two real numbers
        # using `Real` here allows it to work with builtin `int` and `float` types, and numpy scaler types
        if len(values) != 2 or not all(map(lambda v: isinstance(v, Real), values)):
            raise TypeError("limits must be an iterable of two numeric values")
        self._limits = tuple(
            map(round, values)
        )  # if values are close to zero things get weird so round them
        self.selection._limits = self._limits

    @property
    def edge_color(self) -> pygfx.Color:
        """Returns the color of the linear selector."""
        return self._edge_color

    @edge_color.setter
    def edge_color(self, color: str | Sequence[float]):
        """
        Set the color of the linear selector.

        Parameters
        ----------
        color : str | Sequence[float]
            String or sequence of floats that gets converted into a ``pygfx.Color`` object.
        """
        color = pygfx.Color(color)
        # only want to change inner line color
        self._edges[0].material.color = color
        self._original_colors[self._edges[0]] = color
        self._edge_color = color

    # TODO: make `selection` arg in graphics data space not world space
    def __init__(
        self,
        selection: float,
        limits: Sequence[float],
        axis: str = "x",
        parent: Graphic = None,
        edge_color: str | Sequence[float] | np.ndarray = "w",
        thickness: float = 2.5,
        arrow_keys_modifier: str = "Shift",
        name: str = None,
    ):
        """
        Create a horizontal or vertical line that can be used to select a value along an axis.

        Parameters
        ----------
        selection: int
            initial x or y selected position for the selector, in data space

        limits: (int, int)
            (min, max) limits along the x or y-axis for the selector, in data space

        axis: str, default "x"
            "x" | "y", the axis along which the selector can move

        parent: Graphic
            parent graphic for this LinearSelector

        arrow_keys_modifier: str
            modifier key that must be pressed to initiate movement using arrow keys, must be one of:
            "Control", "Shift", "Alt" or ``None``. Double-click the selector first to enable the
            arrow key movements, or set the attribute ``arrow_key_events_enabled = True``

        thickness: float, default 2.5
            thickness of the selector

        edge_color: str | tuple | np.ndarray, default "w"
            color of the selector

        name: str, optional
            name of linear selector

        """
        self._fill_color = None
        self._edge_color = pygfx.Color(edge_color)
        self._vertex_color = None

        if len(limits) != 2:
            raise ValueError("limits must be a tuple of 2 integers, i.e. (int, int)")

        self._limits = np.asarray(limits)

        if axis == "x":
            xs = np.array([selection, selection], dtype=np.float32)
            ys = np.array([0, 1], dtype=np.float32)
            zs = np.zeros(2, dtype=np.float32)

        elif axis == "y":
            xs = np.array([0, 1], dtype=np.float32)
            ys = np.array([selection, selection], dtype=np.float32)
            zs = np.zeros(2, dtype=np.float32)

        else:
            raise ValueError("`axis` must be one of 'x' | 'y'")

        line_data = np.column_stack([xs, ys, zs])

        material = pygfx.LineInfiniteSegmentMaterial

        self.colors_outer = pygfx.Color([0.3, 0.3, 0.3, 1.0])

        line_inner = pygfx.Line(
            # self.data.feature_data because data is a Buffer
            geometry=pygfx.Geometry(positions=line_data),
            material=material(
                thickness=thickness,
                color=edge_color,
                alpha_mode="blend",
                aa=True,
                render_queue=RenderQueue.selector,
                depth_test=False,
                depth_write=False,
                pick_write=True,
            ),
        )

        self.line_outer = pygfx.Line(
            geometry=pygfx.Geometry(positions=line_data),
            material=material(
                thickness=thickness + 6,
                color=self.colors_outer,
                alpha_mode="blend",
                aa=True,
                render_queue=RenderQueue.selector,
                depth_test=False,
                depth_write=False,
                pick_write=True,
            ),
        )

        # Inner line goes on top of the outer line
        line_inner.render_order = 1

        world_object = pygfx.Group()

        world_object.add(self.line_outer)
        world_object.add(line_inner)

        if axis == "x":
            offset = (parent.offset[0], 0, 0)
        elif axis == "y":
            offset = (0, parent.offset[1], 0)

        # init base selector
        BaseSelector.__init__(
            self,
            edges=(line_inner, self.line_outer),
            hover_responsive=(line_inner, self.line_outer),
            arrow_keys_modifier=arrow_keys_modifier,
            axis=axis,
            parent=parent,
            name=name,
            offset=offset,
        )

        self._set_world_object(world_object)

        self._selection = LinearSelectionFeature(
            axis=axis, value=selection, limits=self._limits
        )

        if self._parent is not None:
            self.selection = selection
        else:
            self._selection.set_value(self, selection)

    def get_selected_index(self, graphic: Graphic = None) -> int | list[int]:
        """
        Data index the selector is currently at w.r.t. the Graphic data.

        With LineGraphic data, the geometry x or y position is not always the data position, for example if plotting
        data using np.linspace. Use this to get the data index of the selector.

        Parameters
        ----------
        graphic: Graphic, optional
            Graphic to get the selected data index from. Default is the parent graphic associated to the selector.

        Returns
        -------
        int or List[int]
            data index the selector is currently at, list of ``int`` if a Collection
        """

        source = self._get_source(graphic)

        if isinstance(source, GraphicCollection):
            ixs = list()
            for g in source.graphics:
                ixs.append(self._get_selected_index(g))

            return ixs

        return self._get_selected_index(source)

    def _get_selected_index(self, graphic):
        # the array to search for the closest value along that axis
        if self.axis == "x":
            data = graphic.data[:, 0]
        elif self.axis == "y":
            data = graphic.data[:, 1]

        if (
            "Line" in graphic.__class__.__name__
            or "Scatter" in graphic.__class__.__name__
        ):
            # we want to find the index of the data closest to the selector position
            find_value = self.selection

            # get closest data index to the world space position of the selector
            idx = np.searchsorted(data, find_value, side="left")

            if idx > 0 and (
                idx == len(data)
                or math.fabs(find_value - data[idx - 1])
                < math.fabs(find_value - data[idx])
            ):
                return round(idx - 1)
            else:
                return round(idx)

        if "Image" in graphic.__class__.__name__:
            # indices map directly to grid geometry for image data buffer
            index = self.selection
            shape = graphic.data[:].shape

            if self.axis == "x":
                # assume selecting columns
                upper_bound = shape[1] - 1
            elif self.axis == "y":
                # assume selecting rows
                upper_bound = shape[0] - 1

            return min(round(index), upper_bound)

    def _move_graphic(self, move_info: MoveInfo):
        """
        Moves the graphic

        Parameters
        ----------
        delta: np.ndarray
            delta in world space

        """

        # If this the first move in this drag, store initial selection
        if move_info.start_selection is None:
            move_info.start_selection = self.selection

        delta = move_info.delta[0] if self.axis == "x" else move_info.delta[1]
        self.selection = move_info.start_selection + delta



import ctypes

import numpy as np
import cmap

import wgpu
from imgui_bundle import imgui
from wgpu import GPUTexture

from .. import Popup
from ...utils.functions import (
    COLORMAP_NAMES,
    SEQUENTIAL_CMAPS,
    CYCLIC_CMAPS,
    DIVERGING_CMAPS,
    MISC_CMAPS,
)

all_cmaps = [*SEQUENTIAL_CMAPS, *CYCLIC_CMAPS, *DIVERGING_CMAPS, *MISC_CMAPS]


class ColormapPicker(Popup):
    """Colormap picker menu popup tool"""

    # name used to trigger this popup after it has been registered with a Figure
    name = "colormap-picker"

    def __init__(self, figure):
        super().__init__(figure=figure)

        self.renderer = self._figure.renderer
        self.imgui_renderer = self._figure.imgui_renderer

        # maps str cmap names -> int texture IDs
        self._cmap_texture_refs: dict[str, imgui.ImTextureRef] = dict()

        # make all colormaps and upload representative texture for each cmap to the GPU
        for name in all_cmaps:
            # get data that represents cmap
            colormap = cmap.Colormap(name)
            data = colormap(np.linspace(0, 1)) * 255

            # needs to be 2D to create a texture
            data = np.vstack([[data]] * 2).astype(np.uint8)

            # upload the texture to the GPU, get the texture ID and texture
            self._cmap_texture_refs[name] = self._create_texture_and_upload(data)

        # used to set the states of the UI
        self._lut_tool = None
        self._pos: tuple[int, int] = -1, -1
        self._open_new: bool = False

        self.is_open = False

        self._popup_state = "never-opened"

        self._texture_height = None

    def _create_texture_and_upload(self, data: np.ndarray) -> tuple[int, GPUTexture]:
        """crates a GPUTexture from the 2D data and uploads it"""

        # create a GPUTexture
        texture = self.renderer.device.create_texture(
            size=(data.shape[1], data.shape[0], 4),
            usage=wgpu.TextureUsage.COPY_DST | wgpu.TextureUsage.TEXTURE_BINDING,
            dimension=wgpu.TextureDimension.d2,
            format=wgpu.TextureFormat.rgba8unorm,
            mip_level_count=1,
            sample_count=1,
        )

        # upload to the GPU
        self.renderer.device.queue.write_texture(
            {"texture": texture, "mip_level": 0, "origin": (0, 0, 0)},
            data,
            {"offset": 0, "bytes_per_row": data.shape[1] * 4},
            (data.shape[1], data.shape[0], 1),
        )

        # get a view
        texture_view = texture.create_view()

        # return texture ref
        return self.imgui_renderer.backend.register_texture(texture_view)

    def open(self, pos: tuple[int, int], lut_tool):
        """
        Request that the popup be opened on the next render cycle

        Parameters
        ----------
        pos: int, int
            (x, y) position

        lut_tool: HistogramLUTTool
            instance of the LUT tool

        Returns
        -------

        """
        self._lut_tool = lut_tool

        self._pos = pos

        self._open_new = True

    def close(self):
        """cleanup after popup has closed"""
        self._lut_tool = None
        self._open_new = False
        self._pos = -1, -1

        self.is_open = False

    def _add_cmap_menu_item(self, cmap_name: str):
        # white border around cmap image
        imgui.push_style_color(imgui.Col_.border, (1.0, 1.0, 1.0, 1.0))
        imgui.push_style_var(imgui.StyleVar_.image_border_size, 1.0)

        # cmap image
        texture_ref = self._cmap_texture_refs[cmap_name]
        imgui.image(
            texture_ref,
            image_size=(50, self._texture_height),
        )
        # pop white border
        imgui.pop_style_var()
        imgui.pop_style_color()

        imgui.same_line()

        clicked, selected = imgui.selectable(
            label=cmap_name,
            p_selected=cmap_name == self._lut_tool.cmap,
        )

        if clicked and selected:
            self._lut_tool.cmap = cmap_name

    def update(self):
        if self._open_new:
            # new popup has been triggered by a LUT tool
            self._open_new = False

            imgui.set_next_window_pos(self._pos)
            imgui.open_popup("cmap-picker")

        if imgui.begin_popup("cmap-picker"):
            self.is_open = True

            # make the cmap image height the same as the text height
            self._texture_height = (imgui.get_font_size()) - 2

            if imgui.menu_item("Reset vmin-vmax", "", False)[0]:
                self._lut_tool.images[0].reset_vmin_vmax()

            # add all the cmap options
            for cmap_type in COLORMAP_NAMES.keys():
                if cmap_type == "qualitative":
                    continue

                imgui.separator()
                imgui.text(cmap_type.capitalize())

                for cmap_name in COLORMAP_NAMES[cmap_type]:
                    self._add_cmap_menu_item(cmap_name)

            imgui.end_popup()

        else:
            # popup went from open to closed
            if self.is_open == True:
                self.close()



import math
from typing import *

import pygfx

from ..utils import quick_min_max
from ._base import Graphic
from .selectors import (
    LinearSelector,
    LinearRegionSelector,
    RectangleSelector,
    PolygonSelector,
)
from .features import (
    TextureArray,
    ImageCmap,
    ImageVmin,
    ImageVmax,
    ImageInterpolation,
    ImageCmapInterpolation,
)


class _ImageTile(pygfx.Image):
    """
    Similar to pygfx.Image, only difference is that it modifies the pick_info
    by adding the data row start indices that correspond to this chunk of the big image
    """

    def __init__(
        self,
        geometry,
        material,
        data_slice: tuple[slice, slice],
        chunk_index: tuple[int, int],
        **kwargs,
    ):
        super().__init__(geometry, material, **kwargs)

        self._data_slice = data_slice
        self._chunk_index = chunk_index

    def _wgpu_get_pick_info(self, pick_value):
        pick_info = super()._wgpu_get_pick_info(pick_value)

        data_row_start, data_col_start = (
            self.data_slice[0].start,
            self.data_slice[1].start,
        )

        # add the actual data row and col start indices
        x, y = pick_info["index"]
        x += data_col_start
        y += data_row_start
        pick_info["index"] = (x, y)

        xp, yp = pick_info["pixel_coord"]
        xp += data_col_start
        yp += data_row_start
        pick_info["pixel_coord"] = (xp, yp)

        # add row chunk and col chunk index to pick_info dict
        return {
            **pick_info,
            "data_slice": self.data_slice,
            "chunk_index": self.chunk_index,
        }

    @property
    def data_slice(self) -> tuple[slice, slice]:
        return self._data_slice

    @property
    def chunk_index(self) -> tuple[int, int]:
        return self._chunk_index


class ImageGraphic(Graphic):
    _features = {
        "data": TextureArray,
        "cmap": ImageCmap,
        "vmin": ImageVmin,
        "vmax": ImageVmax,
        "interpolation": ImageInterpolation,
        "cmap_interpolation": ImageCmapInterpolation,
    }

    def __init__(
        self,
        data: Any,
        vmin: float = None,
        vmax: float = None,
        cmap: str = "plasma",
        interpolation: str = "nearest",
        cmap_interpolation: str = "linear",
        isolated_buffer: bool = True,
        **kwargs,
    ):
        """
        Create an Image Graphic

        Parameters
        ----------
        data: array-like
            array-like, usually numpy.ndarray, must support ``memoryview()``
            | shape must be ``[n_rows, n_cols]``, ``[n_rows, n_cols, 3]`` for RGB or ``[n_rows, n_cols, 4]`` for RGBA

        vmin: float, optional
            minimum value for color scaling, estimated from data if not provided

        vmax: float, optional
            maximum value for color scaling, estimated from data if not provided

        cmap: str, optional, default "plasma"
            colormap to use to display the data. For supported colormaps see the
            ``cmap`` library catalogue: https://cmap-docs.readthedocs.io/en/stable/catalog/

        interpolation: str, optional, default "nearest"
            interpolation filter, one of "nearest" or "linear"

        cmap_interpolation: str, optional, default "linear"
            colormap interpolation method, one of "nearest" or "linear"

        isolated_buffer: bool, default True
            If True, initialize a buffer with the same shape as the input data and then
            set the data, useful if the data arrays are ready-only such as memmaps.
            If False, the input array is itself used as the buffer - useful if the
            array is large.

        kwargs:
            additional keyword arguments passed to :class:`.Graphic`

        """

        super().__init__(**kwargs)

        world_object = pygfx.Group()

        if isinstance(data, TextureArray):
            # share buffer
            self._data = data
        else:
            # create new texture array to manage buffer
            # texture array that manages the multiple textures on the GPU that represent this image
            self._data = TextureArray(data, isolated_buffer=isolated_buffer)

        if (vmin is None) or (vmax is None):
            _vmin, _vmax = quick_min_max(self.data.value)
            if vmin is None:
                vmin = _vmin
            if vmax is None:
                vmax = _vmax

        # other graphic features
        self._vmin = ImageVmin(vmin)
        self._vmax = ImageVmax(vmax)

        self._interpolation = ImageInterpolation(interpolation)

        # set map to None for RGB images
        if self._data.value.ndim > 2:
            self._cmap = None
            _map = None
        else:
            # use TextureMap for grayscale images
            self._cmap = ImageCmap(cmap)
            self._cmap_interpolation = ImageCmapInterpolation(cmap_interpolation)

            _map = pygfx.TextureMap(
                self._cmap.texture,
                filter=self._cmap_interpolation.value,
                wrap="clamp-to-edge",
            )

        # one common material is used for every Texture chunk
        self._material = pygfx.ImageBasicMaterial(
            clim=(vmin, vmax),
            map=_map,
            interpolation=self._interpolation.value,
            pick_write=True,
        )

        # iterate through each texture chunk and create
        # an _ImageTile, offset the tile using the data indices
        for texture, chunk_index, data_slice in self._data:
            # create an ImageTile using the texture for this chunk
            img = _ImageTile(
                geometry=pygfx.Geometry(grid=texture),
                material=self._material,
                data_slice=data_slice,  # used to parse pick_info
                chunk_index=chunk_index,
            )

            # row and column start index for this chunk
            data_row_start = data_slice[0].start
            data_col_start = data_slice[1].start

            # offset tile position using the indices from the big data array
            # that correspond to this chunk
            img.world.x = data_col_start
            img.world.y = data_row_start

            world_object.add(img)

        self._set_world_object(world_object)

    @property
    def data(self) -> TextureArray:
        """Get or set the image data"""
        return self._data

    @data.setter
    def data(self, data):
        self._data[:] = data

    @property
    def cmap(self) -> str | None:
        """
        Get or set the colormap for grayscale images. Returns ``None`` if image is RGB(A).

        For supported colormaps see the ``cmap`` library catalogue: https://cmap-docs.readthedocs.io/en/stable/catalog/
        """
        if self._cmap is not None:
            return self._cmap.value

        return None

    @cmap.setter
    def cmap(self, name: str):
        if self.data.value.ndim > 2:
            raise AttributeError("RGB(A) images do not have a colormap property")
        self._cmap.set_value(self, name)

    @property
    def vmin(self) -> float:
        """lower contrast limit"""
        return self._vmin.value

    @vmin.setter
    def vmin(self, value: float):
        self._vmin.set_value(self, value)

    @property
    def vmax(self) -> float:
        """upper contrast limit"""
        return self._vmax.value

    @vmax.setter
    def vmax(self, value: float):
        self._vmax.set_value(self, value)

    @property
    def interpolation(self) -> str:
        """Data interpolation method"""
        return self._interpolation.value

    @interpolation.setter
    def interpolation(self, value: str):
        self._interpolation.set_value(self, value)

    @property
    def cmap_interpolation(self) -> str:
        """cmap interpolation method"""
        return self._cmap_interpolation.value

    @cmap_interpolation.setter
    def cmap_interpolation(self, value: str):
        self._cmap_interpolation.set_value(self, value)

    def reset_vmin_vmax(self):
        """
        Reset the vmin, vmax by estimating it from the data by subsampling.
        """

        vmin, vmax = quick_min_max(self._data.value)
        self.vmin = vmin
        self.vmax = vmax

    def add_linear_selector(
        self, selection: int = None, axis: str = "x", **kwargs
    ) -> LinearSelector:
        """
        Adds a :class:`.LinearSelector`.

        Selectors are just ``Graphic`` objects, so you can manage, remove, or delete them
        from a plot area just like any other ``Graphic``.

        Parameters
        ----------
        selection: int, optional
            initial position of the selector

        kwargs:
            passed to :class:`.LinearSelector`

        Returns
        -------
        LinearSelector

        """

        if axis == "x":
            limits = (0, self._data.value.shape[1])
        elif axis == "y":
            limits = (0, self._data.value.shape[0])
        else:
            raise ValueError("`axis` must be one of 'x' | 'y'")

        if selection is None:
            selection = limits[0]

        if selection < limits[0] or selection > limits[1]:
            raise ValueError(
                f"the passed selection: {selection} is beyond the limits: {limits}"
            )

        selector = LinearSelector(
            selection=selection,
            limits=limits,
            axis=axis,
            parent=self,
            **kwargs,
        )

        self._plot_area.add_graphic(selector, center=False)

        return selector

    def add_linear_region_selector(
        self,
        selection: tuple[float, float] = None,
        axis: str = "x",
        padding: float = 0.0,
        fill_color=(0, 0, 0.35, 0.2),
        **kwargs,
    ) -> LinearRegionSelector:
        """
        Add a :class:`.LinearRegionSelector`.

        Selectors are just ``Graphic`` objects, so you can manage, remove, or delete them
        from a plot area just like any other ``Graphic``.

        Parameters
        ----------
        selection: (float, float)
            initial (min, max) of the selection

        axis: "x" | "y"
            axis the selector can move along

        padding: float, default 100.0
            Extends the linear selector along the perpendicular axis to make it easier to interact with.

        kwargs
            passed to ``LinearRegionSelector``

        Returns
        -------
        LinearRegionSelector

        """

        if axis == "x":
            size = self._data.value.shape[0]
            center = size / 2
            limits = (0, self._data.value.shape[1])
        elif axis == "y":
            size = self._data.value.shape[1]
            center = size / 2
            limits = (0, self._data.value.shape[0])
        else:
            raise ValueError("`axis` must be one of 'x' | 'y'")

        # default padding is 25% the height or width of the image
        if padding is None:
            size *= 1.25
        else:
            size += padding

        if selection is None:
            selection = limits[0], int(limits[1] * 0.25)

        if padding is None:
            size *= 1.25

        else:
            size += padding

        selector = LinearRegionSelector(
            selection=selection,
            limits=limits,
            size=size,
            center=center,
            axis=axis,
            fill_color=fill_color,
            parent=self,
            **kwargs,
        )

        self._plot_area.add_graphic(selector, center=False)

        return selector

    def add_rectangle_selector(
        self,
        selection: tuple[float, float, float, float] = None,
        fill_color=(0, 0, 0.35, 0.2),
        **kwargs,
    ) -> RectangleSelector:
        """
        Add a :class:`.RectangleSelector`.

        Selectors are just ``Graphic`` objects, so you can manage, remove, or delete them
        from a plot area just like any other ``Graphic``.

        Parameters
        ----------
        selection: (float, float, float, float), optional
            initial (xmin, xmax, ymin, ymax) of the selection

        """
        # default selection is 25% of the diagonal
        if selection is None:
            diagonal = math.sqrt(
                self._data.value.shape[0] ** 2 + self._data.value.shape[1] ** 2
            )

            selection = (0, int(diagonal / 4), 0, int(diagonal / 4))

        # min/max limits are image shape
        # rows are ys, columns are xs
        limits = (0, self._data.value.shape[1], 0, self._data.value.shape[0])

        selector = RectangleSelector(
            selection=selection,
            limits=limits,
            fill_color=fill_color,
            parent=self,
            **kwargs,
        )

        self._plot_area.add_graphic(selector, center=False)

        return selector

    def add_polygon_selector(
        self,
        selection: List[tuple[float, float]] = None,
        fill_color=(0, 0, 0.35, 0.2),
        **kwargs,
    ) -> PolygonSelector:
        """
        Add a :class:`.PolygonSelector`.

        Selectors are just ``Graphic`` objects, so you can manage, remove, or delete them
        from a plot area just like any other ``Graphic``.

        Parameters
        ----------
        selection: List of positions, optional
            Initial points for the polygon. If not given or None, you'll start drawing the selection (clicking adds points to the polygon).

        """

        # min/max limits are image shape
        # rows are ys, columns are xs
        limits = (0, self._data.value.shape[1], 0, self._data.value.shape[0])

        selector = PolygonSelector(
            selection,
            limits,
            fill_color=fill_color,
            parent=self,
            **kwargs,
        )

        self._plot_area.add_graphic(selector, center=False)

        return selector



import pygfx
import numpy as np

from ..utils.enums import RenderQueue
from ._base import Graphic
from .features import (
    TextData,
    FontSize,
    TextFaceColor,
    TextOutlineColor,
    TextOutlineThickness,
)


class TextGraphic(Graphic):
    _features = {
        "text": TextData,
        "font_size": FontSize,
        "face_color": TextFaceColor,
        "outline_color": TextOutlineColor,
        "outline_thickness": TextOutlineThickness,
    }

    def __init__(
        self,
        text: str,
        font_size: float | int = 14,
        face_color: str | np.ndarray | list[float] | tuple[float] = "w",
        outline_color: str | np.ndarray | list[float] | tuple[float] = "w",
        outline_thickness: float = 0.0,
        screen_space: bool = True,
        offset: tuple[float] = (0, 0, 0),
        anchor: str = "middle-center",
        **kwargs,
    ):
        """
        Create a text Graphic

        Parameters
        ----------
        text: str
            text to display

        font_size: float | int, default 10
            font size

        face_color: str, array, list, tuple, default "w"
            str or RGBA array to set the color of the text

        outline_color: str, array, list, tuple, default "w"
            str or RGBA array to set the outline color of the text

        outline_thickness: float, default 0
            relative outline thickness, value between 0.0 - 0.5

        screen_space: bool = True
            if True, text size is in screen space, if False the text size is in data space

        offset: (float, float, float), default (0, 0, 0)
            places the text at this location

        anchor: str, default "middle-center"
            position of the origin of the text
            a string representing the vertical and horizontal anchors, separated by a dash

            * Vertical values: "top", "middle", "baseline", "bottom"
            * Horizontal values: "left", "center", "right"

        **kwargs
            passed to :class:`.Graphic`

        """

        super().__init__(**kwargs)

        self._text = TextData(text)
        self._font_size = FontSize(font_size)
        self._face_color = TextFaceColor(face_color)
        self._outline_color = TextOutlineColor(outline_color)
        self._outline_thickness = TextOutlineThickness(outline_thickness)

        # Text is usually used for annotations and the like. But we still want it to write depth.
        # We make it render later than other 'auto' objects, assuming that most of these don't have transparent fragments.

        # The aa is on because it makes the glyphs prettier. It can result in artifacts, but these often express a different color outline
        # which is actually not so bad; it would look weird on a line, but for text it helps the contrast of the glyph!

        world_object = pygfx.Text(
            text=self.text,
            font_size=self.font_size,
            screen_space=screen_space,
            anchor=anchor,
            material=pygfx.TextMaterial(
                alpha_mode="auto",
                render_queue=RenderQueue.auto + 50,
                aa=True,
                color=self.face_color,
                outline_color=self.outline_color,
                outline_thickness=self.outline_thickness,
                pick_write=True,
            ),
        )

        self._set_world_object(world_object)

        self.offset = offset

    @property
    def world_object(self) -> pygfx.Text:
        """Text world object"""
        return super(TextGraphic, self).world_object

    @property
    def text(self) -> str:
        """Get or set the text"""
        return self._text.value

    @text.setter
    def text(self, text: str):
        self._text.set_value(self, text)

    @property
    def font_size(self) -> float | int:
        """Get or set the font size"""
        return self._font_size.value

    @font_size.setter
    def font_size(self, size: float | int):
        self._font_size.set_value(self, size)

    @property
    def face_color(self) -> pygfx.Color:
        """Get or set the face color"""
        return self._face_color.value

    @face_color.setter
    def face_color(self, color: str | np.ndarray | list[float] | tuple[float]):
        self._face_color.set_value(self, color)

    @property
    def outline_thickness(self) -> float:
        """Get or set the outline thickness"""
        return self._outline_thickness.value

    @outline_thickness.setter
    def outline_thickness(self, thickness: float):
        self._outline_thickness.set_value(self, thickness)

    @property
    def outline_color(self) -> pygfx.Color:
        """Get or set the outline color"""
        return self._outline_color.value

    @outline_color.setter
    def outline_color(self, color: str | np.ndarray | list[float] | tuple[float]):
        self._outline_color.set_value(self, color)



from numbers import Real
from typing import Sequence

import numpy as np
import pygfx

from .._base import Graphic
from .._collection_base import GraphicCollection
from ..features._selection_features import LinearRegionSelectionFeature
from ._base_selector import BaseSelector, MoveInfo
from ...utils.enums import RenderQueue


class LinearRegionSelector(BaseSelector):
    _features = {"selection": LinearRegionSelectionFeature}

    @property
    def parent(self) -> Graphic | None:
        """graphic that the selector is associated with"""
        return self._parent

    @property
    def selection(self) -> np.ndarray[float]:
        """
        (min, max) of selector along selector's axis
        """
        # TODO: This probably does not account for rotation since world.position
        #  does not account for rotation, we can do this later

        return self._selection.value.copy()

        # TODO: if no parent graphic is set, this just returns values in world space
        #  but should we change it?

    @selection.setter
    def selection(self, selection: Sequence[float]):
        # set (min, max) of the selector in data space
        graphic = self._parent

        self._selection.set_value(self, selection)

    @property
    def limits(self) -> tuple[float, float]:
        return self._limits

    @limits.setter
    def limits(self, values: tuple[float, float]):
        # check that `values` is an iterable of two real numbers
        # using `Real` here allows it to work with builtin `int` and `float` types, and numpy scaler types
        if len(values) != 2 or not all(map(lambda v: isinstance(v, Real), values)):
            raise TypeError("limits must be an iterable of two numeric values")
        self._limits = tuple(
            map(round, values)
        )  # if values are close to zero things get weird so round them
        self._selection._limits = self._limits

    def __init__(
        self,
        selection: Sequence[float],
        limits: Sequence[float],
        size: float,
        center: float,
        axis: str = "x",
        parent: Graphic = None,
        resizable: bool = True,
        fill_color: str | Sequence[float] = (0, 0, 0.35),
        edge_color: str | Sequence[float] = (0.8, 0.6, 0),
        edge_thickness: float = 8,
        arrow_keys_modifier: str = "Shift",
        name: str = None,
    ):
        """
        Create a LinearRegionSelector graphic which can be moved only along either the x-axis or y-axis.
        Allows sub-selecting data from a parent ``Graphic`` or from multiple Graphics.

        Assumes that the data under the selector is a function of the axis on which the selector moves
        along. Example: if the selector is along the x-axis, then there must be only one y-value for each
        x-value, otherwise methods such as ``get_selected_data()`` do not make sense.

        Parameters
        ----------
        selection: (float, float)
            initial (min, max) x or y values

        limits: (float, float)
            (min limit, max limit) within which the selector can move

        size: int
            usually the data range, height or width of the selector

        center: float
            usually the data mean, center offset of the selector on the orthogonal axis

        axis: str, default "x"
            "x" | "y", axis along which the selector can move

        parent: ``Graphic`` instance, default ``None``
            associate this selector with a parent ``Graphic`` from which to fetch data or indices

        resizable: bool
            if ``True``, the edges can be dragged to change the range of the selection

        fill_color: str, array, or tuple
            fill color for the selector, passed to pygfx.Color

        edge_color: str, array, or tuple
            edge color for the selector, passed to pygfx.Color

        edge_thickness: float, default 8
            edge thickness

        arrow_keys_modifier: str
            modifier key that must be pressed to initiate movement using arrow keys, must be one of:
            "Control", "Shift", "Alt" or ``None``

        name: str, optional
            name of this selector graphic

        """
        self._edge_color = pygfx.Color(edge_color)
        self._fill_color = pygfx.Color(fill_color)
        self._vertex_color = None

        # lots of very close to zero values etc. so round them, otherwise things get weird
        if not len(selection) == 2:
            raise ValueError

        selection = np.asarray(selection)

        if not len(limits) == 2:
            raise ValueError

        self._limits = np.asarray(limits)

        # TODO: sanity checks, we recommend users to add LinearSelection using the add_linear_selector() methods
        # TODO: so we can worry about the sanity checks later

        group = pygfx.Group()

        if axis == "x":
            mesh = pygfx.Mesh(
                pygfx.box_geometry(1, size, 1),
                pygfx.MeshBasicMaterial(
                    color=pygfx.Color(self.fill_color),
                    alpha_mode="blend",
                    opacity=0.4,
                    render_queue=RenderQueue.selector,
                    depth_test=False,
                    depth_write=False,
                    pick_write=True,
                ),
            )

        elif axis == "y":
            mesh = pygfx.Mesh(
                pygfx.box_geometry(size, 1, 1),
                pygfx.MeshBasicMaterial(
                    color=pygfx.Color(self.fill_color),
                    alpha_mode="blend",
                    opacity=0.4,
                    render_queue=RenderQueue.selector,
                    depth_test=False,
                    depth_write=False,
                    pick_write=True,
                ),
            )
        else:
            raise ValueError("`axis` must be one of 'x' or 'y'")

        # Render the mesh before the lines
        mesh.render_order = -1

        # the fill of the selection
        self.fill = mesh
        # no x, y offsets for linear region selector
        # everything is done by setting the mesh data
        # and line positions
        self.fill.world.position = (0, 0, -2)

        group.add(self.fill)

        self._resizable = resizable

        if axis == "x":
            # just some data to initialize the edge lines
            init_line_data = np.array([[0, -size / 2, 0], [0, size / 2, 0]]).astype(
                np.float32
            )

        elif axis == "y":
            # just some line data to initialize y axis edge lines
            init_line_data = np.array(
                [
                    [-size / 2, 0, 0],
                    [size / 2, 0, 0],
                ]
            ).astype(np.float32)

        else:
            raise ValueError("axis argument must be one of 'x' or 'y'")

        line0 = pygfx.Line(
            pygfx.Geometry(
                positions=init_line_data.copy()
            ),  # copy so the line buffer is isolated
            pygfx.LineMaterial(
                thickness=edge_thickness,
                color=self.edge_color,
                alpha_mode="blend",
                opacity=1,
                aa=True,
                render_queue=RenderQueue.selector,
                depth_test=False,
                depth_write=False,
                pick_write=True,
            ),
        )
        line1 = pygfx.Line(
            pygfx.Geometry(
                positions=init_line_data.copy()
            ),  # copy so the line buffer is isolated
            pygfx.LineMaterial(
                thickness=edge_thickness,
                color=self.edge_color,
                alpha_mode="blend",
                opacity=1,
                aa=True,
                render_queue=RenderQueue.selector,
                depth_test=False,
                depth_write=False,
                pick_write=True,
            ),
        )

        self.edges: tuple[pygfx.Line, pygfx.Line] = (line0, line1)
        group.add(*self.edges)

        # TODO: if parent offset changes, we should set the selector offset too, use offset evented property
        # TODO: add check if parent is `None`, will throw error otherwise
        if axis == "x":
            offset = (parent.offset[0], center + parent.offset[1], 0)
        elif axis == "y":
            offset = (center + parent.offset[1], parent.offset[1], 0)

        # set the initial bounds of the selector
        # compensate for any offset from the parent graphic
        # selection feature only works in world space, not data space
        self._selection = LinearRegionSelectionFeature(
            selection, axis=axis, limits=self._limits
        )

        self._pygfx_event = None

        BaseSelector.__init__(
            self,
            edges=self.edges,
            fill=(self.fill,),
            hover_responsive=self.edges,
            arrow_keys_modifier=arrow_keys_modifier,
            axis=axis,
            parent=parent,
            name=name,
            offset=offset,
        )

        self._set_world_object(group)

        self.selection = selection

    def get_selected_data(
        self, graphic: Graphic = None
    ) -> np.ndarray | list[np.ndarray]:
        """
        Get the ``Graphic`` data bounded by the current selection.
        Returns a view of the data array.

        If the ``Graphic`` is a collection, such as a ``LineStack``, it returns a list of views of the full array.
        Can be performed on the ``parent`` Graphic or on another graphic by passing to the ``graphic`` arg.

        **NOTE:** You must be aware of the axis for the selector. The sub-selected data that is returned will be of
        shape ``[n_points_selected, 3]``. If you have selected along the x-axis then you can access y-values of the
        subselection like this: sub[:, 1]. Conversely, if you have selected along the y-axis then you can access the
        x-values of the subselection like this: sub[:, 0].

        Parameters
        ----------
        graphic: Graphic, optional, default ``None``
            if provided, returns the data selection from this graphic instead of the graphic set as ``parent``

        Returns
        -------
        np.ndarray or List[np.ndarray]
            view or list of views of the full array, returns ``None`` if selection is empty

        """

        source = self._get_source(graphic)
        ixs = self.get_selected_indices(source)

        if "Line" in source.__class__.__name__:
            if isinstance(source, GraphicCollection):
                # this will return a list of views of the arrays, therefore no copy operations occur
                # it's fine and fast even as a list of views because there is no re-allocating of memory
                # this is fast even for slicing a 10,000 x 5,000 LineStack
                data_selections: list[np.ndarray] = list()

                for i, g in enumerate(source.graphics):
                    if ixs[i].size == 0:
                        data_selections.append(
                            np.array([], dtype=np.float32).reshape(0, 3)
                        )
                    else:
                        s = slice(
                            ixs[i][0], ixs[i][-1] + 1
                        )  # add 1 because these are direct indices
                        # slices n_datapoints dim
                        data_selections.append(g.data[s])

                return data_selections
            else:
                if ixs.size == 0:
                    # empty selection
                    return np.array([], dtype=np.float32).reshape(0, 3)

                s = slice(
                    ixs[0], ixs[-1] + 1
                )  # add 1 to end because these are direct indices
                # slices n_datapoints dim
                # slice with min, max is faster than using all the indices
                return source.data[s]

        if "Image" in source.__class__.__name__:
            s = slice(ixs[0], ixs[-1] + 1)

            if self.axis == "x":
                # slice columns
                return source.data[:, s]

            elif self.axis == "y":
                # slice rows
                return source.data[s]

    def get_selected_indices(
        self, graphic: Graphic = None
    ) -> np.ndarray | list[np.ndarray]:
        """
        Returns the indices of the ``Graphic`` data bounded by the current selection.

        These are the data indices along the selector's "axis" which correspond to the data under the selector.

        Parameters
        ----------
        graphic: Graphic, optional
            if provided, returns the selection indices from this graphic instead of the graphic set as ``parent``

        Returns
        -------
        Union[np.ndarray, List[np.ndarray]]
            data indices of the selection, list of np.ndarray if graphic is a collection

        """
        # we get the indices from the source graphic
        source = self._get_source(graphic)

        # get the offset of the source graphic
        if self.axis == "x":
            dim = 0
        elif self.axis == "y":
            dim = 1

        # selector (min, max) data values along axis
        bounds = self.selection

        if (
            "Line" in source.__class__.__name__
            or "Scatter" in source.__class__.__name__
        ):
            # gets indices corresponding to n_datapoints dim
            # data is [n_datapoints, xyz], so we return
            # indices that can be used to slice `n_datapoints`
            if isinstance(source, GraphicCollection):
                ixs = list()
                for g in source.graphics:
                    # indices for each graphic in the collection
                    data = g.data[:, dim]
                    g_ixs = np.where((data >= bounds[0]) & (data <= bounds[1]))[0]
                    ixs.append(g_ixs)
            else:
                # map this only this graphic
                data = source.data[:, dim]
                ixs = np.where((data >= bounds[0]) & (data <= bounds[1]))[0]

            return ixs

        if "Image" in source.__class__.__name__:
            # indices map directly to grid geometry for image data buffer
            return np.arange(*bounds, dtype=int)

    def _move_graphic(self, move_info: MoveInfo):
        # If this the first move in this drag, store initial selection
        if move_info.start_selection is None:
            move_info.start_selection = self.selection

        # add delta to current min, max to get new positions
        delta = move_info.delta[0] if self.axis == "x" else move_info.delta[1]

        # Get original selection
        cur_min, cur_max = move_info.start_selection

        # move entire selector if event source was fill
        if move_info.source == self.fill:
            # Limit the delta to avoid weird resizine behavior
            min_delta = self.limits[0] - cur_min
            max_delta = self.limits[1] - cur_max
            delta = np.clip(delta, min_delta, max_delta)
            # Update both bounds with equal amount
            self._selection.set_value(self, (cur_min + delta, cur_max + delta))
            return

        # if selector not resizable return
        if not self._resizable:
            return

        # if event source was an edge and selector is resizable,
        # move the edge that caused the event
        if move_info.source == self.edges[0]:
            # change only left or bottom bound
            new_min = min(cur_min + delta, cur_max)
            self._selection.set_value(self, (new_min, cur_max))

        elif move_info.source == self.edges[1]:
            # change only right or top bound
            new_max = max(cur_max + delta, cur_min)
            self._selection.set_value(self, (cur_min, new_max))



from functools import partial
from collections import OrderedDict
from typing import Iterable

import numpy as np
import pygfx

from ..utils.enums import RenderQueue
from ..graphics import Graphic
from ..graphics.features import GraphicFeatureEvent
from ..graphics import LineGraphic, ScatterGraphic, ImageGraphic
from ..utils import mesh_masks


class LegendItem:
    def __init__(
        self,
        label: str,
        color: pygfx.Color,
    ):
        """

        Parameters
        ----------
        label: str

        color: pygfx.Color
        """
        self._label = label
        self._color = color


class LineLegendItem(LegendItem):
    def __init__(
        self, parent, graphic: LineGraphic, label: str, position: tuple[int, int]
    ):
        """

        Parameters
        ----------
        graphic: LineGraphic

        label: str

        position: [x, y]
        """

        if label is not None:
            pass

        elif graphic.name is not None:
            pass

        else:
            raise ValueError(
                "Must specify `label` or Graphic must have a `name` to auto-use as the label"
            )

        # for now only support lines with a single color
        if np.unique(graphic.colors.value, axis=0).shape[0] > 1:
            raise ValueError("Use colorbars for multi-colored lines, not legends")

        color = pygfx.Color(np.unique(graphic.colors.value, axis=0).ravel())

        self._parent = parent

        super().__init__(label, color)

        graphic.colors.add_event_handler(self._update_color)

        # construct Line WorldObject
        data = np.array([[0, 0, 0], [3, 0, 0]], dtype=np.float32)

        self._line_world_object = pygfx.Line(
            geometry=pygfx.Geometry(positions=data),
            material=pygfx.LineMaterial(
                alpha_mode="blend",
                render_queue=RenderQueue.overlay,
                thickness=8,
                color=self._color,
                depth_write=False,
                depth_test=False,
            ),
        )

        # self._line_world_object.world.x = position[0]

        self._label_world_object = pygfx.Text(
            text=str(label),
            font_size=6,
            screen_space=False,
            anchor="middle-left",
            material=pygfx.TextMaterial(
                alpha_mode="blend",
                aa=True,
                render_queue=RenderQueue.overlay,
                color="w",
                outline_color="w",
                outline_thickness=0,
                depth_write=False,
                depth_test=False,
            ),
        )

        self.world_object = pygfx.Group()
        self.world_object.add(self._line_world_object, self._label_world_object)

        self.world_object.world.x = position[0]
        # add 10 to x to account for space for the line
        self._label_world_object.world.x = position[0] + 10

        self.world_object.world.y = position[1]

        self.world_object.add_event_handler(
            partial(self._highlight_graphic, graphic), "click"
        )

    @property
    def label(self) -> str:
        return self._label

    @label.setter
    def label(self, text: str):
        self._parent._check_label_unique(text)
        self._label_world_object.geometry.set_text(text)

    def _update_color(self, ev: GraphicFeatureEvent):
        new_color = ev.info["value"]
        if np.unique(new_color, axis=0).shape[0] > 1:
            raise ValueError(
                "LegendError: LineGraphic colors no longer appropriate for legend"
            )

        self._color = new_color[0]
        self._line_world_object.material.color = pygfx.Color(self._color)

    def _highlight_graphic(self, graphic: Graphic, ev):
        graphic_color = pygfx.Color(np.unique(graphic.colors.value, axis=0).ravel())

        if graphic_color == self._parent.highlight_color:
            graphic.colors = self._color
        else:
            # hacky but fine for now
            orig_color = pygfx.Color(self._color)
            graphic.colors = self._parent.highlight_color
            self._color = orig_color


class Legend(Graphic):
    def __init__(
        self,
        plot_area,
        highlight_color: str | tuple | np.ndarray = "w",
        max_rows: int = 5,
        *args,
        **kwargs,
    ):
        """

        Parameters
        ----------
        plot_area: Union[Plot, Subplot, Dock]
            plot area to put the legend in

        highlight_color: Union[str, tuple, np.ndarray], default "w"
            highlight color

        max_rows: int, default 5
            maximum number of rows allowed in the legend

        """
        self._graphics: list[Graphic] = list()

        # hex id of Graphic, i.e. graphic._fpl_address are the keys
        self._items: OrderedDict[str:LegendItem] = OrderedDict()

        super().__init__(*args, **kwargs)

        group = pygfx.Group()
        self._legend_items_group = pygfx.Group()
        self._set_world_object(group)

        self._mesh = pygfx.Mesh(
            pygfx.box_geometry(50, 10, 1),
            pygfx.MeshBasicMaterial(
                alpha_mode="blend",
                render_queue=RenderQueue.overlay,
                color=pygfx.Color([0.1, 0.1, 0.1, 1]),
                wireframe_thickness=10,
                depth_write=False,
                depth_test=False,
            ),
        )

        # Plane gets rendered before text and line
        self._mesh.render_order = -1

        self.world_object.add(self._mesh)
        self.world_object.add(self._legend_items_group)

        self.highlight_color = pygfx.Color(highlight_color)

        self._plot_area = plot_area
        self._plot_area.add_graphic(self)

        if self._plot_area.__class__.__name__ == "Dock":
            if self._plot_area.size < 1:
                self._plot_area.size = 100

        # TODO: refactor with "moveable graphic" base class once that's done
        self._mesh.add_event_handler(self._pointer_down, "pointer_down")
        self._plot_area.renderer.add_event_handler(self._pointer_move, "pointer_move")
        self._plot_area.renderer.add_event_handler(self._pointer_up, "pointer_up")

        self._last_position = None
        self._initial_controller_state = self._plot_area.controller.enabled

        self._max_rows = max_rows

        self._row_counter = 0
        self._col_counter = 0

    def graphics(self) -> tuple[Graphic, ...]:
        return tuple(self._graphics)

    def _check_label_unique(self, label):
        for legend_item in self._items.values():
            if legend_item.label == label:
                raise ValueError(
                    f"You have passed the label '{label}' which is already used for another legend item. "
                    f"All labels within a legend must be unique."
                )

    def add_graphic(self, graphic: Graphic, label: str = None):
        if graphic in self._graphics:
            raise KeyError(
                f"Graphic already exists in legend with label: '{self._items[graphic._fpl_address].label}'"
            )

        self._check_label_unique(label)

        new_col_ix = self._col_counter
        new_row_ix = self._row_counter

        x_pos = 0
        y_pos = 0

        if self._row_counter == self._max_rows:
            # set counters
            new_col_ix = self._col_counter + 1

            # get x position offset for this new column of LegendItems
            # start by getting the LegendItems in the previous column
            prev_column_items: list[LegendItem] = list(self._items.values())[
                -self._max_rows :
            ]
            # x position of LegendItems in previous column
            x_pos = prev_column_items[-1].world_object.world.x
            max_width = 0
            # get width of widest LegendItem in previous column to add to x_pos offset for this column
            for item in prev_column_items:
                bbox = item.world_object.get_world_bounding_box()
                width, height, depth = np.ptp(bbox, axis=0)
                max_width = max(max_width, width)

            # x position offset for this new column
            x_pos = x_pos + max_width + 15  # add 15 for spacing

            # rest row index for next iteration
            new_row_ix = 1
        else:
            if len(self._items) > 0:
                x_pos = list(self._items.values())[-1].world_object.world.x

            y_pos = new_row_ix * -10
            new_row_ix = self._row_counter + 1

        if isinstance(graphic, LineGraphic):
            legend_item = LineLegendItem(self, graphic, label, position=(x_pos, y_pos))
        else:
            raise ValueError("Legend only supported for LineGraphic for now.")

        self._legend_items_group.add(legend_item.world_object)
        self._reset_mesh_dims()

        self._graphics.append(graphic)
        self._items[graphic._fpl_address] = legend_item

        graphic.add_event_handler(partial(self.remove_graphic, graphic), "deleted")

        self._col_counter = new_col_ix
        self._row_counter = new_row_ix

    def _reset_mesh_dims(self):
        bbox = self._legend_items_group.get_world_bounding_box()

        width, height, _ = np.ptp(bbox, axis=0)

        self._mesh.geometry.positions.data[mesh_masks.x_right] = width + 7
        self._mesh.geometry.positions.data[mesh_masks.x_left] = -5
        self._mesh.geometry.positions.data[mesh_masks.y_bottom] = 0
        self._mesh.geometry.positions.data[mesh_masks.y_bottom] = -height - 3
        self._mesh.geometry.positions.update_range()

    def remove_graphic(self, graphic: Graphic):
        self._graphics.remove(graphic)
        legend_item = self._items.pop(graphic._fpl_address)
        self._legend_items_group.remove(legend_item.world_object)
        self._reset_item_positions()

    def _reset_item_positions(self):
        for i, (graphic_loc, legend_item) in enumerate(self._items.items()):
            y_pos = i * -10
            legend_item.world_object.world.y = y_pos

        self._reset_mesh_dims()

    def reorder(self, labels: Iterable[str]):
        all_labels = [legend_item.label for legend_item in self._items.values()]

        if not set(labels) == set(all_labels):
            raise ValueError("Must pass all existing legend labels")

        new_items = OrderedDict()

        for label in labels:
            for graphic_loc, legend_item in self._items.items():
                if label == legend_item.label:
                    new_items[graphic_loc] = self._items.pop(graphic_loc)
                    break

        self._items = new_items
        self._reset_item_positions()

    def _pointer_down(self, ev):
        self._last_position = self._plot_area.map_screen_to_world(ev)
        self._initial_controller_state = self._plot_area.controller.enabled

    def _pointer_move(self, ev):
        if self._last_position is None:
            return

        self._plot_area.controller.enabled = False

        world_pos = self._plot_area.map_screen_to_world(ev)

        # outside viewport
        if world_pos is None:
            return

        delta = world_pos - self._last_position

        self.world_object.world.x = self.world_object.world.x + delta[0]
        self.world_object.world.y = self.world_object.world.y + delta[1]

        self._last_position = world_pos

        self._plot_area.controller.enabled = self._initial_controller_state

    def _pointer_up(self, ev):
        self._last_position = None
        if self._initial_controller_state is not None:
            self._plot_area.controller.enabled = self._initial_controller_state

    def __getitem__(self, graphic: Graphic) -> LegendItem:
        if not isinstance(graphic, Graphic):
            raise TypeError("Must index Legend with Graphics")

        if graphic._fpl_address not in self._items.keys():
            raise KeyError("Graphic not in legend")

        return self._items[graphic._fpl_address]



from typing import Sequence

import numpy as np
import pygfx as gfx

from ...utils import mesh_masks
from ._base import GraphicFeature, GraphicFeatureEvent, block_reentrance
from ...utils.triangulation import triangulate


class LinearSelectionFeature(GraphicFeature):
    event_info_spec = [
        {
            "dict key": "value",
            "type": "float",
            "description": "new x or y value of selection",
        },
    ]

    event_extra_attrs = [
        {
            "attribute": "get_selected_index",
            "type": "callable",
            "description": "returns index under the selector",
        }
    ]

    def __init__(self, axis: str, value: float, limits: tuple[float, float]):
        """

        Parameters
        ----------
        axis: "x" | "y"
            axis the selector is restricted to

        value: float
            position of the slider in world space, NOT data space
        limits: (float, float)
            min, max limits of the selector

        """

        super().__init__()

        self._axis = axis
        self._limits = limits
        self._value = value

    @property
    def value(self) -> np.float32:
        """
        selection, data x or y value
        """
        return self._value

    @block_reentrance
    def set_value(self, selector, value: float):
        # clip value between limits
        value = np.clip(value, self._limits[0], self._limits[1], dtype=np.float32)

        # set position
        if self._axis == "x":
            dim = 0
        elif self._axis == "y":
            dim = 1

        for edge in selector._edges:
            edge.geometry.positions.data[:, dim] = value
            edge.geometry.positions.update_range()

        self._value = value

        event = GraphicFeatureEvent("selection", {"value": value})
        event.get_selected_index = selector.get_selected_index

        self._call_event_handlers(event)


class LinearRegionSelectionFeature(GraphicFeature):
    event_info_spec = [
        {
            "dict key": "value",
            "type": "np.ndarray",
            "description": "new [min, max] of selection",
        },
    ]

    event_extra_attrs = [
        {
            "attribute": "get_selected_indices",
            "type": "callable",
            "description": "returns indices under the selector",
        },
        {
            "attribute": "get_selected_data",
            "type": "callable",
            "description": "returns data under the selector",
        },
    ]

    def __init__(self, value: tuple[int, int], axis: str, limits: tuple[float, float]):
        super().__init__()

        self._axis = axis
        self._limits = limits
        self._value = tuple(int(v) for v in value)

    @property
    def value(self) -> np.ndarray[float]:
        """
        (min, max) of the selection, in data space
        """
        return self._value

    @property
    def axis(self) -> str:
        """one of "x" | "y" """
        return self._axis

    @block_reentrance
    def set_value(self, selector, value: Sequence[float]):
        """
        Set start, stop range of selector

        Parameters
        ----------
        selector: LinearRegionSelector

        value: (float, float)
             (min, max) values in data space

        """
        if not len(value) == 2:
            raise TypeError(
                "selection must be a array, tuple, list, or sequence in the form of `(min, max)`, "
                "where `min` and `max` are numeric values."
            )

        # convert to array, clip values if they are beyond the limits
        value = np.asarray(value, dtype=np.float32).clip(*self._limits)

        # make sure `selector width >= 2`, left edge must not move past right edge!
        # or bottom edge must not move past top edge!
        if not (value[1] - value[0]) >= 0:
            return

        if self.axis == "x":
            # change left x position of the fill mesh
            selector.fill.geometry.positions.data[mesh_masks.x_left] = value[0]

            # change right x position of the fill mesh
            selector.fill.geometry.positions.data[mesh_masks.x_right] = value[1]

            # change x position of the left edge line
            selector.edges[0].geometry.positions.data[:, 0] = value[0]

            # change x position of the right edge line
            selector.edges[1].geometry.positions.data[:, 0] = value[1]

        elif self.axis == "y":
            # change bottom y position of the fill mesh
            selector.fill.geometry.positions.data[mesh_masks.y_bottom] = value[0]

            # change top position of the fill mesh
            selector.fill.geometry.positions.data[mesh_masks.y_top] = value[1]

            # change y position of the bottom edge line
            selector.edges[0].geometry.positions.data[:, 1] = value[0]

            # change y position of the top edge line
            selector.edges[1].geometry.positions.data[:, 1] = value[1]

        self._value = value

        # send changes to GPU
        selector.fill.geometry.positions.update_range()

        selector.edges[0].geometry.positions.update_range()
        selector.edges[1].geometry.positions.update_range()

        # send event
        if len(self._event_handlers) < 1:
            return

        event = GraphicFeatureEvent("selection", {"value": self.value})

        event.get_selected_indices = selector.get_selected_indices
        event.get_selected_data = selector.get_selected_data

        self._call_event_handlers(event)
        # TODO: user's selector event handlers can call event.graphic.get_selected_indices() to get the data index,
        #  and event.graphic.get_selected_data() to get the data under the selection
        #  this is probably a good idea so that the data isn't sliced until it's actually necessary


class RectangleSelectionFeature(GraphicFeature):
    event_info_spec = [
        {
            "dict key": "value",
            "type": "np.ndarray",
            "description": "new [xmin, xmax, ymin, ymax] of selection",
        },
    ]

    event_extra_attrs = [
        {
            "attribute": "get_selected_indices",
            "type": "callable",
            "description": "returns indices under the selector",
        },
        {
            "attribute": "get_selected_data",
            "type": "callable",
            "description": "returns data under the selector",
        },
    ]

    def __init__(
        self,
        value: tuple[float, float, float, float],
        limits: tuple[float, float, float, float],
    ):
        super().__init__()

        self._limits = limits
        self._value = tuple(int(v) for v in value)

    @property
    def value(self) -> np.ndarray[float]:
        """
        (xmin, xmax, ymin, ymax) of the selection, in data space
        """
        return self._value

    @block_reentrance
    def set_value(self, selector, value: Sequence[float]):
        """
        Set the selection of the rectangle selector.

        Parameters
        ----------
        selector: RectangleSelector

        value: (float, float, float, float)
            new values (xmin, xmax, ymin, ymax) of the selection
        """
        if not len(value) == 4:
            raise TypeError(
                "Selection must be an array, tuple, list, or sequence in the form of `(xmin, xmax, ymin, ymax)`, "
                "where `xmin`, `xmax`, `ymin`, `ymax` are numeric values."
            )

        # convert to array
        value = np.asarray(value, dtype=np.float32)

        # clip values if they are beyond the limits
        value[:2] = value[:2].clip(self._limits[0], self._limits[1])
        # clip y
        value[2:] = value[2:].clip(self._limits[2], self._limits[3])

        xmin, xmax, ymin, ymax = value

        # make sure `selector width >= 2` and selector height >=2 , left edge must not move past right edge!
        # or bottom edge must not move past top edge!
        if not (xmax - xmin) >= 0 or not (ymax - ymin) >= 0:
            return

        # change fill mesh
        # change left x position of the fill mesh
        selector.fill.geometry.positions.data[mesh_masks.x_left] = xmin

        # change right x position of the fill mesh
        selector.fill.geometry.positions.data[mesh_masks.x_right] = xmax

        # change bottom y position of the fill mesh
        selector.fill.geometry.positions.data[mesh_masks.y_bottom] = ymin

        # change top position of the fill mesh
        selector.fill.geometry.positions.data[mesh_masks.y_top] = ymax

        # change the edge lines

        # each edge line is defined by two end points which are stored in the
        # geometry.positions
        # [x0, y0, z0]
        # [x1, y1, z0]

        # left line
        z = selector.edges[0].geometry.positions.data[:, -1][0]
        selector.edges[0].geometry.positions.data[:] = np.array(
            [[xmin, ymin, z], [xmin, ymax, z]]
        )

        # right line
        selector.edges[1].geometry.positions.data[:] = np.array(
            [[xmax, ymin, z], [xmax, ymax, z]]
        )

        # bottom line
        selector.edges[2].geometry.positions.data[:] = np.array(
            [[xmin, ymin, z], [xmax, ymin, z]]
        )

        # top line
        selector.edges[3].geometry.positions.data[:] = np.array(
            [[xmin, ymax, z], [xmax, ymax, z]]
        )

        # change the vertex positions

        # bottom left
        selector.vertices[0].geometry.positions.data[:] = np.array([[xmin, ymin, 1]])

        # bottom right
        selector.vertices[1].geometry.positions.data[:] = np.array([[xmax, ymin, 1]])

        # top left
        selector.vertices[2].geometry.positions.data[:] = np.array([[xmin, ymax, 1]])

        # top right
        selector.vertices[3].geometry.positions.data[:] = np.array([[xmax, ymax, 1]])

        self._value = value

        # send changes to GPU
        selector.fill.geometry.positions.update_range()

        for edge in selector.edges:
            edge.geometry.positions.update_range()

        for vertex in selector.vertices:
            vertex.geometry.positions.update_range()

        # send event
        if len(self._event_handlers) < 1:
            return

        event = GraphicFeatureEvent("selection", {"value": self.value})

        event.get_selected_indices = selector.get_selected_indices
        event.get_selected_data = selector.get_selected_data

        # calls any events
        self._call_event_handlers(event)


class PolygonSelectionFeature(GraphicFeature):
    event_info_spec = [
        {
            "dict key": "value",
            "type": "np.ndarray",
            "description": "new array of points that represents the polygon selection",
        },
    ]

    event_extra_attrs = [
        {
            "attribute": "get_selected_indices",
            "type": "callable",
            "description": "returns indices under the selector",
        },
        {
            "attribute": "get_selected_data",
            "type": "callable",
            "description": "returns data under the selector",
        },
    ]

    def __init__(
        self,
        value: Sequence[tuple[float]],
        limits: tuple[float, float, float, float],
    ):
        super().__init__()

        self._limits = limits
        self._value = np.asarray(value).reshape(-1, 3).astype(float)

    @property
    def value(self) -> np.ndarray[float]:
        """
        The array of the polygon, in data space
        """
        return self._value

    @block_reentrance
    def set_value(self, selector, value: Sequence[tuple[float]]):
        """
        Set the selection of the rectangle selector.

        Parameters
        ----------
        selector: PolygonSelector

        value: array
            new values (3D points) of the selection
        """

        value = np.asarray(value, dtype=np.float32)

        if not value.shape[1] == 3:
            raise TypeError(
                "Selection must be an array, tuple, list, or sequence of the shape Nx3."
            )

        # clip values if they are beyond the limits
        value[:, 0] = value[:, 0].clip(self._limits[0], self._limits[1])
        value[:, 1] = value[:, 1].clip(self._limits[2], self._limits[3])

        self._value = value

        if len(value) >= 3:
            indices = triangulate(value)
        else:
            indices = np.zeros((0, 3), np.int32)

        geometry = selector.geometry

        # Need larger buffer?
        if len(value) > geometry.positions.nitems:
            arr = np.zeros((geometry.positions.nitems * 2, 3), np.float32)
            geometry.positions = gfx.Buffer(arr)
        if len(indices) > geometry.indices.nitems:
            arr = np.zeros((geometry.indices.nitems * 2, 3), np.int32)
            geometry.indices = gfx.Buffer(arr)

        geometry.positions.data[: len(value)] = value
        geometry.positions.data[len(value) :] = value[-1] if len(value) else (0, 0, 0)
        geometry.positions.draw_range = 0, len(value)
        geometry.positions.update_full()

        geometry.indices.data[: len(indices)] = indices
        geometry.indices.data[len(indices) :] = 0
        geometry.indices.draw_range = 0, len(indices)
        geometry.indices.update_full()

        # send event
        if len(self._event_handlers) < 1:
            return

        event = GraphicFeatureEvent("selection", {"value": self.value})

        event.get_selected_indices = selector.get_selected_indices
        event.get_selected_data = selector.get_selected_data

        # calls any events
        self._call_event_handlers(event)
